# -*- coding: utf-8 -*-
"""
Created on Sat Aug 12 16:08:41 2023

@author: ZHANG Jun
"""

import os
from .exceptions import FileExit
from .incar_tag import INCAR_TAG

def generate_file_name(fname):
    while os.path.exists(fname):
        fname = fname + '_new'
    return fname

def file_exit():
    if os.path.exists('StopPython'):
        os.remove('StopPython')
        raise FileExit('Exit because `StopPython` file is found.')

def modify_INCAR(key='NSW', value='300', s=''):
    """Modify the INCAR file.

    :param key: The INCAR tag, defaults to 'NSW'
    :type key: str, optional
    :param value: Value of the INCAR tag, defaults to '300'
    :type value: str, optional
    :param s: Comment string, defaults to ''
    :type s: str, optional
    :return: Modified INCAR file

    """

    if not key in INCAR_TAG:
        print('Input key not avaliable, please check.')
        return 1

    new_incar, discover_code = [], False
    with open('INCAR', 'r') as f:
        for line in f:
            str_list = line.split()
            if len(str_list) == 0:
                new_incar.append('\n')
            elif str_list[0] == key:
                str_list[2] = value
                new_incar.append(f'  {str_list[0]} = {str_list[2]}\n')
                discover_code = True
            else:
                new_incar.append(line)

    if s:
        new_incar.append(f'\n{s}\n')

    if not discover_code:
        new_incar.append(f'  {key} = {value}\n')

    with open('INCAR', 'w') as f:
        for line in new_incar:
            f.write(line)
    return 0



# This script is used to find atoms connected with specified atomic index.
# from pymatgen.core.structure import Structure # pymatgen module deprecated for now.
import numpy as np
import sys
from collections import Counter

def find_adsorption_site():
    """

    :return: DESCRIPTION
    :rtype: TYPE

    """
    from pymatgen.core.structure import Structure

    assert len(sys.argv) > 3, 'Usage: command + file name + atomic index (start from zero) + distance cutoff'
    fname = sys.argv[1]
    atom_num = sys.argv[2]
    dist_cutoff = sys.argv[3]

    crystal = Structure.from_file(fname)
    element_list = [crystal.sites[x].specie.name for x in range(crystal.num_sites)]
    dist_mat = crystal.distance_matrix[int(atom_num)]
    OneNN_index = np.where(dist_mat < float(dist_cutoff))

    # sort_index = sorted(range(len(dist_mat)), key=lambda k: dist_mat[k])
    ele = np.array(element_list)[OneNN_index]
    for i in ['O', 'H']:
        location = np.where(ele == i)
        ele = np.delete(ele, location)
    for i in ele:
        print(i, end='  ')
    print()



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

# import important object.
from .find_adsorption_site import find_adsorption_site
from .generate_adsorption_sites import AddAtoms
from .high_throughput_predict import HtAds
from .increase_vacuum import increase_vacuum
from .volcano_2e import volcano_2e
from .volcano_plot import volcano_plot

from .high_throughput_dft_calculation import HtDftAds




# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

from .adsorbate_poscar import adsorbate_poscar
from .exceptions import FileExit
from .file_lib import generate_file_name, file_exit, modify_INCAR
from .model_lib import save_model, load_model, save_state_dict, load_state_dict
from .model_lib import config_parser, EarlyStopping, load_graph_build_method, PearsonR




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 12 16:08:41 2023

@author: ZHANG Jun
"""
import os
class FileExit(Exception):
    pass

def file_exit():
    if os.path.exists('StopPython'):
        os.remove('StopPython')
        raise FileExit('Exit because `StopPython` file is found.')




# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

from .layer import Layer
from .model import PotentialModel, CrystalPropertyModel, AtomicPropertyModel, AtomicVectorModel
from .fit import Fit


# -*- coding: utf-8 -*-
"""
Created on Sun Sep 17 18:17:36 2023

@author: ZHANG Jun
"""

import platform
import numpy as np
import os
import stat
import shutil

from ase.formula import Formula
from ase.io import read
from ase.lattice.cubic import FaceCenteredCubic
from ase.data import covalent_radii, atomic_numbers

def perturb_positions(atoms, amplitude=0.1):
    calculator  = atoms.get_calculator()
    atoms       = atoms.copy()
    atoms.set_calculator(calculator)
    posistions  = atoms.arrays['positions']
    num_atoms   = len(atoms)
    increment   = np.clip(np.random.normal(0.0, amplitude / 3,  size=(num_atoms,3)), -amplitude, amplitude) # np.random.uniform(-amplitude, amplitude, (num_atoms,3))
    constraints = atoms.constraints
    if len(constraints) > 0:
        increment[constraints[0].index] = 0.0
    posistions += increment
    atoms.set_positions(posistions)
    return atoms

def scale_atoms(atoms, scale_factor=1.0):
    calculator = atoms.get_calculator()
    new_atoms  = atoms.copy()
    new_atoms.set_calculator(calculator)
    cell        = new_atoms.get_cell()
    frac_coords = new_atoms.get_scaled_positions()
    new_cell    = cell * scale_factor
    new_atoms.set_cell(new_cell)
    new_atoms.set_scaled_positions(frac_coords)
    return new_atoms

def get_concentration_from_ase_formula(formula):
    f_dict = Formula(formula).count()
    tot = np.sum(list(f_dict.values()))
    c_dict = {k: v/tot for k, v in f_dict.items()}
    c_dict = {k:c_dict[k] for k in c_dict if c_dict[k] > 0}
    return c_dict

def get_v_per_atom(chemical_formula):
    frac_dict = get_concentration_from_ase_formula(chemical_formula)
    frac_dict = {k:(0.0 if k not in frac_dict else frac_dict[k]) for k in ['Ni', 'Co', 'Fe', 'Pd', 'Pt']}
    return -282.7957531391954 * (frac_dict['Ni'] + frac_dict['Co'] + frac_dict['Fe'])\
        - 278.79605077419797 * frac_dict['Pd'] - 278.6228860885035 * frac_dict['Pt']\
            + 293.66128761358624

def get_ase_atom_from_formula(chemical_formula, v_per_atom=None):
    # interpret formula
    atomic_fracions = get_concentration_from_ase_formula(chemical_formula)
    elements = [x for x in atomic_fracions]
    element_number = [atomic_numbers[x] for x in elements]
    mean_radii       = np.sum([covalent_radii[n] * atomic_fracions[e] for n, e in zip(element_number, elements)]) #

    Pt_radii           = covalent_radii[78]
    latticeconstant    = mean_radii / Pt_radii * 3.92
    atoms              = FaceCenteredCubic('Pt', directions=[[1,-1,0], [0,1,-1], [1,1,1]], size=(4, 4, 2), latticeconstant=latticeconstant, pbc=True)
    total_atom         = len(atoms)
    num_atom_list      = np.array(list(atomic_fracions.values())) * total_atom
    num_atom_list      = np.around(num_atom_list, decimals=0)
    total_tmp          = np.sum(num_atom_list)
    deviation          = total_atom - total_tmp
    num_atom_list[np.random.randint(len(elements))] += deviation

    # shuffle atoms
    ase_number    = []
    for i_index, i in enumerate(num_atom_list):
        for j in range(int(i)):
            ase_number.append(element_number[i_index])
    np.random.shuffle(ase_number)
    atoms.set_atomic_numbers(ase_number)

    # scale atoms
    if isinstance(v_per_atom, (float, int)):
        volume = atoms.cell.volume
        volume_per_atom = volume / len(atoms)
        volume_ratio = v_per_atom / volume_per_atom
        scale_factor = pow(volume_ratio, 1/3)
        atoms = scale_atoms(atoms, scale_factor)
    return atoms

def get_ase_atom_from_formula_template( chemical_formula, v_per_atom=None,
                                       template_file='POSCAR_temp'):
    # interpret formula
    # the template file should be a bulk structure
    atomic_fracions    = get_concentration_from_ase_formula(chemical_formula)
    elements           = [x for x in atomic_fracions]
    element_number     = [atomic_numbers[x] for x in elements]
    atoms              = read(template_file)
    total_atom         = len(atoms)
    num_atom_list      = np.array(list(atomic_fracions.values())) * total_atom
    num_atom_list      = np.around(num_atom_list, decimals=0)
    total_tmp          = np.sum(num_atom_list)
    deviation          = total_atom - total_tmp
    num_atom_list[np.random.randint(len(elements))] += deviation

    # shuffle atoms
    ase_number    = []
    for i_index, i in enumerate(num_atom_list):
        for j in range(int(i)):
            ase_number.append(element_number[i_index])
    np.random.shuffle(ase_number)
    atoms.set_atomic_numbers(ase_number)

    # scale atoms
    if isinstance(v_per_atom, (float, int)):
        volume = atoms.cell.volume
        volume_per_atom = volume / len(atoms)
        volume_ratio = v_per_atom / volume_per_atom
        scale_factor = pow(volume_ratio, 1/3)
        atoms = scale_atoms(atoms, scale_factor)
    return atoms

def run_vasp(vasp_bash_path):
    """

    :raises ValueError: VASP can only run on a Linux platform


    .. warning:: Setup your own VAPS package and Intel libraries before using this function.

    """

    # os_type = platform.system()
    # if not os_type == 'Linux':
    #     raise ValueError(f'VASP can only be executed on Linux OS, instead of {os_type}.')
    # shell_script = '''#!/bin/bash
# . /home/jzhang/software/intel/oneapi/setvars.sh
# mpirun /home/jzhang/software/vasp/vasp_std
    # '''

    # with open('vasp_run.sh', 'w') as f:
    #     f.write(shell_script)
    # 
    # os.chmod('vasp_run.sh', stat.S_IRWXU)

    shutil.copyfile(vasp_bash_path, os.path.join('./vasp_run.sh'))
    os.chmod('vasp_run.sh', stat.S_IRWXU)
    os.system('./vasp_run.sh')

# remove imported objects
# del np, Formula, read, FaceCenteredCubic, covalent_radii, atomic_numbers, platform



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 22:49:14 2023

@author: ZHANG Jun
"""

import numpy as np
import os
import time

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, random_split
from torch.optim import lr_scheduler

from .model import PotentialModel
from ..lib.model_lib import EarlyStopping, PearsonR, config_parser, save_model, load_state_dict, save_state_dict
from ..data.load_dataset import LoadDataset, Collater
from ..default_parameters import default_train_config
from ..lib.file_lib import file_exit

class Fit(object):
    def __init__(self, **train_config):
        self.train_config = {**default_train_config, **config_parser(train_config)}
        self.log = open('fit.log', 'w', buffering=1)

        # check device
        self.device = self.train_config['device']
        self.verbose = self.train_config['verbose']
        if torch.cuda.is_available() and self.device == 'cpu':
            print('User warning: `CUDA` device is available, but you choosed `cpu`.', file=self.log)
        elif not torch.cuda.is_available() and self.device.split(':')[0] == 'cuda':
            print('User warning: `CUDA` device is not available, but you choosed \
`cuda:0`. Change the device to `cpu`.', file=self.log)
            self.device = 'cpu'
        print('User info: Specified device for potential model:', self.device, file=self.log)

        # read dataset
        print(f'User info: Loading dataset from {self.train_config["dataset_path"]}',
              file=self.log)
        self._dataset=LoadDataset(self.train_config['dataset_path'])

        # split dataset
        self._val_size = int(len(self._dataset)*self.train_config['validation_size'])
        self._test_size = int(len(self._dataset)*self.train_config['test_size'])
        self._train_size = len(self._dataset) - self._val_size - self._test_size
        train_dataset, val_dataset, test_dataset = random_split(self._dataset,
                                                                [self._train_size,
                                                                 self._val_size,
                                                                 self._test_size])
        # check batch size
        self.train_config['batch_size'] = min(self.train_config['batch_size'],
                                              self._train_size)
        self.train_config['val_batch_size'] = min(self.train_config['val_batch_size'],
                                                  self._val_size,
                                                  self._test_size)

        # instantiate data loader
        collate_fn = Collater(device=self.device)
        self.train_loader = DataLoader(train_dataset,
                                       batch_size=self.train_config['batch_size'],
                                       shuffle=True,
                                       num_workers=0,
                                       collate_fn=collate_fn)
        self.val_loader = DataLoader(val_dataset,
                                     batch_size=self.train_config['val_batch_size'],
                                     shuffle=True,
                                     num_workers=0,
                                     collate_fn=collate_fn)
        self.test_loader = DataLoader(test_dataset,
                                      batch_size=self.train_config['val_batch_size'],
                                      shuffle=True,
                                      num_workers=0,
                                      collate_fn=collate_fn)

        # check the existence of adsorbates.
        _graph, _ = self._dataset[0]
        self._has_adsorbate = _graph.ndata.__contains__('adsorbate')

        # check neural nodes dimensions, and modify them if necessary.
        num_heads = len(self.train_config['head_list'])
        atomic_depth = self._dataset[0][0].ndata['h'].size()[1]
        if self.train_config['gat_node_dim_list'][0] != atomic_depth:
            print(f'User warning: Input dimension of the first AGAT `Layer` \
(the first element of `gat_node_dim_list`) should equal to the dimension of \
atomic representation depth. Your input is: {self.train_config["gat_node_dim_list"][0]}, \
which is changed to be: `{atomic_depth}`.', file=self.log)
            self.train_config['gat_node_dim_list'][0] = atomic_depth
        in_dim = num_heads * self.train_config['gat_node_dim_list'][-1]
        if self.train_config['energy_readout_node_list'][0] != in_dim:
            print(f"User warning: Input dimension of the first energy readout layer \
(the first element of `energy_readout_node_list`) should equal to \
`len(self.train_config['head_list']) * self.train_config['gat_node_dim_list'][-1]`. \
Your input is: {self.train_config['energy_readout_node_list'][0]}, which is \
changed to be: `{in_dim}`.", file=self.log)
            self.train_config['energy_readout_node_list'][0] = in_dim
        if self.train_config['energy_readout_node_list'][-1] != 1:
            print(f"User warning: Output dimension of the last energy readout layer \
(the last element of `energy_readout_node_list`) should equal to \
`1`. Your input is: {self.train_config['energy_readout_node_list'][-1]}, which is \
changed to be: `1`.", file=self.log)
            self.train_config['energy_readout_node_list'][-1] = 1
        if self.train_config['force_readout_node_list'][0] != in_dim:
            print(f"User warning: Input dimension of the first force readout layer \
(the first element of `force_readout_node_list`) should equal to \
`len(self.train_config['head_list']) * self.train_config['gat_node_dim_list'][-1]`. \
Your input is: {self.train_config['force_readout_node_list'][0]}, which is \
changed to be: `{in_dim}`.", file=self.log)
            self.train_config['force_readout_node_list'][0] = in_dim
        if self.train_config['force_readout_node_list'][-1] != 3:
            print(f"User warning: Output dimension of the last force readout layer \
(the last element of `force_readout_node_list`) should equal to \
`3`. Your input is: {self.train_config['force_readout_node_list'][-1]}, which is \
changed to be: `3`.", file=self.log)
            self.train_config['force_readout_node_list'][-1] = 3
        if self.train_config['stress_readout_node_list'][0] != in_dim:
            print(f"User warning: Input dimension of the first stress readout layer \
(the first element of `stress_readout_node_list`) should equal to \
`len(self.train_config['head_list']) * self.train_config['gat_node_dim_list'][-1]`. \
Your input is: {self.train_config['stress_readout_node_list'][0]}, which is \
changed to be: `{in_dim}`.", file=self.log)
            self.train_config['stress_readout_node_list'][0] = in_dim
        if self.train_config['stress_readout_node_list'][-1] != 6:
            print(f"User warning: Output dimension of the last stress readout layer \
(the last element of `stress_readout_node_list`) should equal to \
`6`. Your input is: {self.train_config['stress_readout_node_list'][-1]}, which is \
changed to be: `6`.", file=self.log)
            self.train_config['stress_readout_node_list'][-1] = 6

        # prepare out file
        if not os.path.exists(self.train_config['output_files']):
            os.mkdir(self.train_config['output_files'])

    def fit(self, **train_config):
        # update config if needed.
        self.train_config = {**self.train_config, **config_parser(train_config)}

        # construct a model and an optimizer.
        model = PotentialModel(self.train_config['gat_node_dim_list'],
                               self.train_config['energy_readout_node_list'],
                               self.train_config['force_readout_node_list'],
                               self.train_config['stress_readout_node_list'],
                               self.train_config['head_list'],
                               self.train_config['bias'],
                               self.train_config['negative_slope'],
                               self.device,
                               self.train_config['tail_readout_no_act']
                               )

        optimizer = optim.Adam(model.parameters(),
                              lr=self.train_config['learning_rate'],
                              weight_decay=self.train_config['weight_decay'])

        # load stat dict if there exists.
        if os.path.exists(os.path.join(self.train_config['model_save_dir'],
                                       'agat_state_dict.pth')):
            try:
                checkpoint = load_state_dict(self.train_config['model_save_dir'])
                model.load_state_dict(checkpoint['model_state_dict'])
                model.eval()
                model = model.to(self.device)
                model.device = self.device
                optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
                print(f'User info: Model and optimizer state dict loaded successfully from {self.train_config["model_save_dir"]}.', file=self.log)
            except:
                print('User warning: Exception catched when loading model and optimizer state dict.', file=self.log)
        else:
            print('User info: Checkpoint not detected', file=self.log)

        # transfer learning.
        if self.train_config['transfer_learning']:
            for param in model.gat_layers.parameters():
                param.requires_grad = False

            exp_lr_scheduler = lr_scheduler.StepLR(optimizer, step_size=100,
                                                   gamma=0.1, verbose=True)

        # early stop
        if self.train_config['early_stop']:
            stopper = EarlyStopping(model, self.log,
                                    patience=self.train_config['stop_patience'],
                                    model_save_dir=self.train_config['model_save_dir'])

        # loss function
        criterion = self.train_config['criterion']
        a, b, c = self.train_config['a'], self.train_config['b'], self.train_config['c']
        mae = nn.L1Loss()
        r = PearsonR

        # log file
        print('========================================================================', file=self.log)
        print(model, file=self.log)
        print('========================================================================', file=self.log)
        if self.verbose > 1:
            print("Epoch Batch Energy_Loss Force_Loss Stress_Loss Total_Loss Dur_(s) Train_info",
                  file=self.log)
        if self.verbose > 0:
            print("Epoch Energy_Loss Force_Loss Stress_Loss Total_Loss \
Energy_MAE Force_MAE Stress_MAE Energy_R Force_R Stress_R Dur_(s) Validation_info",
              file=self.log)

        # start the training
        start_time= time.time()
        for epoch in range(self.train_config['epochs']):
            # release GPU memory
            # torch.cuda.empty_cache()

            for i, (graph, props) in enumerate(self.train_loader):
                energy_true = props['energy_true']
                force_true = graph.ndata['forces_true']
                stress_true = props['stress_true']
                optimizer.zero_grad()
                energy_pred, force_pred, stress_pred = model.forward(graph)
                energy_loss = criterion(energy_pred, energy_true)
                if self._has_adsorbate:
                    force_true *= torch.reshape(graph.ndata['adsorbate']*self.train_config['adsorbate_coeff']+1.,
                                            (-1,1))
                    force_pred *= torch.reshape(graph.ndata['adsorbate']*self.train_config['adsorbate_coeff']+1.,
                                            (-1,1))

                # if self._has_adsorbate:
                #     force_loss = ...
                force_loss = criterion(force_pred, force_true)
                stress_loss = criterion(stress_pred, stress_true)
                total_loss = a*energy_loss + b*force_loss + c*stress_loss
                total_loss.backward()
                optimizer.step()
                dur = time.time() - start_time
                if self.verbose > 1:
                    print("{:0>5d} {:0>5d} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:10.1f} Train_info".format(
                          epoch, i, energy_loss.item(), force_loss.item(),
                          stress_loss.item(), total_loss.item(), dur),
                          file=self.log)

            # validation every epoch
            with torch.no_grad():
                energy_true_all, force_true_all, stress_true_all = [], [], []
                energy_pred_all, force_pred_all, stress_pred_all = [], [], []
                for i, (graph, props) in enumerate(self.val_loader):
                    energy_true_all.append(props['energy_true'])
                    force_true = graph.ndata['forces_true']
                    force_true *= torch.reshape(graph.ndata['adsorbate']*self.train_config['adsorbate_coeff']+1.,
                                                (-1,1))
                    force_true_all.append(force_true)
                    stress_true_all.append(props['stress_true'])
                    energy_pred, force_pred, stress_pred = model.forward(graph)
                    energy_pred_all.append(energy_pred)
                    force_pred *= torch.reshape(graph.ndata['adsorbate']*self.train_config['adsorbate_coeff']+1.,
                                                (-1,1))
                    force_pred_all.append(force_pred)
                    stress_pred_all.append(stress_pred)

                energy_true_all = torch.cat(energy_true_all)
                energy_pred_all = torch.cat(energy_pred_all)
                force_true_all = torch.cat(force_true_all)
                force_pred_all = torch.cat(force_pred_all)
                stress_true_all = torch.cat(stress_true_all)
                stress_pred_all = torch.cat(stress_pred_all)
                energy_loss = criterion(energy_pred_all, energy_true_all)
                force_loss = criterion(force_pred_all, force_true_all)
                stress_loss = criterion(stress_pred_all, stress_true_all)
                total_loss = a*energy_loss + b*force_loss + c*stress_loss

                energy_mae = mae(energy_pred_all, energy_true_all)
                force_mae = mae(force_pred_all, force_true_all)
                stress_mae = mae(stress_pred_all, stress_true_all)

                energy_r = r(energy_pred_all, energy_true_all)
                force_r = r(force_pred_all, force_true_all)
                stress_r = r(stress_pred_all, stress_true_all)
                if self.verbose > 0:
                    print("{:0>5d} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:1.8f} {:10.1f} Validation_info".format(
                          epoch, energy_loss.item(), force_loss.item(), stress_loss.item(),
                          total_loss.item(), energy_mae.item(), force_mae.item(),
                          stress_mae.item(), energy_r.item(), force_r.item(),
                          stress_r.item(), dur),
                          file=self.log)

            if self.train_config['early_stop']:
                if stopper.step(total_loss, epoch, model, optimizer):
                    break
                if stopper.update:
                    energy = torch.cat([torch.reshape(energy_pred_all, (-1,1)),
                                        torch.reshape(energy_true_all, (-1,1))],
                                       dim=1).cpu().numpy()
                    force = torch.cat([torch.reshape(force_pred_all, (-1,1)),
                                       torch.reshape(force_true_all, (-1,1))],
                                      dim=1).cpu().numpy()
                    stress = torch.cat([torch.reshape(stress_pred_all, (-1,1)),
                                        torch.reshape(stress_true_all, (-1,1))],
                                       dim=1).cpu().numpy()
                    np.savetxt(os.path.join(self.train_config['output_files'],
                                            'energy_val_pred_true.txt'),
                               energy, fmt='%.8f')
                    np.savetxt(os.path.join(self.train_config['output_files'],
                                            'force_val_pred_true.txt'),
                               force, fmt='%.8f')
                    np.savetxt(os.path.join(self.train_config['output_files'],
                                            'stress_val_pred_true.txt'),
                               stress, fmt='%.8f')
            else:
                save_model(model, model_save_dir=self.train_config['model_save_dir'])
                save_state_dict(model, state_dict_save_dir=self.train_config['model_save_dir'])

            if self.train_config['transfer_learning']:
                exp_lr_scheduler.step()
            file_exit()

        # test with the best model
        try:
            checkpoint = load_state_dict(self.train_config['model_save_dir'])
            model.load_state_dict(checkpoint['model_state_dict'])
            model.eval()
            model = model.to(self.device)
            model.device = self.device
            print(f'User info: Test model state dict loaded successfully from {self.train_config["model_save_dir"]}.', file=self.log)
        except:
            print('User warning: Exception catched when loading test model state dict. \nUsing train model instead.', file=self.log)

        with torch.no_grad():
            energy_true_all, force_true_all, stress_true_all = [], [], []
            energy_pred_all, force_pred_all, stress_pred_all = [], [], []
            for i, (graph, props) in enumerate(self.test_loader):
                energy_true_all.append(props['energy_true'])
                force_true_all.append(graph.ndata['forces_true'])
                stress_true_all.append(props['stress_true'])
                energy_pred, force_pred, stress_pred = model.forward(graph)
                energy_pred_all.append(energy_pred)
                force_pred_all.append(force_pred)
                stress_pred_all.append(stress_pred)

            energy_true_all = torch.cat(energy_true_all)
            energy_pred_all = torch.cat(energy_pred_all)
            force_true_all = torch.cat(force_true_all)
            force_pred_all = torch.cat(force_pred_all)
            stress_true_all = torch.cat(stress_true_all)
            stress_pred_all = torch.cat(stress_pred_all)

            energy_loss = criterion(energy_pred_all, energy_true_all)
            force_loss = criterion(force_pred_all, force_true_all)
            stress_loss = criterion(stress_pred_all, stress_true_all)
            total_loss = a*energy_loss + b*force_loss + c*stress_loss

            energy_mae = mae(energy_pred_all, energy_true_all)
            force_mae = mae(force_pred_all, force_true_all)
            stress_mae = mae(stress_pred_all, stress_true_all)

            energy_r = r(energy_pred_all, energy_true_all)
            force_r = r(force_pred_all, force_true_all)
            stress_r = r(stress_pred_all, stress_true_all)

            print(f'''User info, model performance on testset: (No sample weight on the loss)
    Epoch      : {epoch}
    Energy loss: {energy_loss.item()}
    Force_Loss : {force_loss.item()}
    Stress_Loss: {stress_loss.item()}
    Total_Loss : {total_loss.item()}
    Energy_MAE : {energy_mae.item()}
    Force_MAE  : {force_mae.item()}
    Stress_MAE : {stress_mae.item()}
    Energy_R   : {energy_r.item()}
    Force_R    : {force_r.item()}
    Stress_R   : {stress_r.item()}
    Dur (s)    : {dur}''',
            file=self.log)

            energy = torch.cat([torch.reshape(energy_pred_all, (-1,1)),
                                torch.reshape(energy_true_all, (-1,1))],
                               dim=1).cpu().numpy()
            force = torch.cat([torch.reshape(force_pred_all, (-1,1)),
                               torch.reshape(force_true_all, (-1,1))],
                              dim=1).cpu().numpy()
            stress = torch.cat([torch.reshape(stress_pred_all, (-1,1)),
                                torch.reshape(stress_true_all, (-1,1))],
                               dim=1).cpu().numpy()
            np.savetxt(os.path.join(self.train_config['output_files'],
                                    'energy_test_pred_true.txt'),
                       energy, fmt='%.8f')
            np.savetxt(os.path.join(self.train_config['output_files'],
                                    'force_test_pred_true.txt'),
                       force, fmt='%.8f')
            np.savetxt(os.path.join(self.train_config['output_files'],
                                    'stress_test_pred_true.txt'),
                       stress, fmt='%.8f')

        self.log.close()
        return total_loss

if __name__ == '__main__':
    FIX_VALUE = [1,3,6]
    train_config = {
        'verbose': 1,
    'dataset_path': os.path.join('dataset', 'concated_graphs.bin'),
    'model_save_dir': 'agat_model',
    'epochs': 1000,
    'output_files': 'out_file',
    'device': 'cuda:0',
    'validation_size': 0.15,
    'test_size': 0.15,
    'early_stop': True,
    'stop_patience': 300,
    'gat_node_dim_list': [6, 100, 100, 100],
    'head_list': ['mul', 'div', 'free'],
    'energy_readout_node_list': [300, 100, 50, 30, 10, 3, FIX_VALUE[0]],
    'force_readout_node_list': [300, 100, 50, 30, 10, FIX_VALUE[1]],
    'stress_readout_node_list': [300, FIX_VALUE[2]],
    'bias': True,
    'negative_slope': 0.2,
    'criterion': nn.MSELoss(),
    'a': 1.0,
    'b': 1.0,
    'c': 0.0,
    'optimizer': 'adam', # Fix to sgd.
    'learning_rate': 0.0005,
    'weight_decay': 0.0, # weight decay (L2 penalty)
    'batch_size': 64,
    'val_batch_size': 400,
    # 'validation_batches': 150,
    'transfer_learning': False,
    'trainable_layers': -4,
    'mask_fixed': False,
    'tail_readout_no_act': [3,3,1],
    'adsorbate': False, # indentify adsorbate or not when building graphs.
    'adsorbate_coeff': 20.0, # the importance of adsorbate atoms with respective to surface atoms.
    'transfer_learning': False}

    f = Fit(**train_config)
    f.fit()



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 13 17:12:52 2023

@author: 18326
"""

import numpy as np
import os
import shutil
import sys

from ase.build import add_vacuum, sort
from ase.constraints import FixAtoms
from ase.io import read

from .generate_adsorption_sites import AddAtoms
from ...lib.model_lib import config_parser
from ...lib.adsorbate_poscar import adsorbate_poscar
from ...default_parameters import default_hp_dft_config
from ...lib.file_lib import modify_INCAR
from ...lib.high_throughput_lib import get_ase_atom_from_formula, get_v_per_atom, run_vasp

class HtDftAds(object):
    """High-throughput DFT calculations for adsorption. See details in: https://jzhang-github.github.io/AGAT/Default%20parameters.html#default_hp_dft_config

    .. py:property:: root_dir

       The root working directory of this object.

    :param **hp_config: Configurations to control the process.
    :type **hp_config: dict

    Example::

        HA = HpDftAds(calculation_index=0)
        HA.run('NiCoFePdPt')
    """
    def __init__(self, **hp_config):
        self.hp_config = {**default_hp_dft_config, **config_parser(hp_config)}
        self.root_dir = os.getcwd()

    def bulk_opt(self, formula):
        """Structural optimization of the bulk structure with VASP.

        :param formula: Chemical formula
        :type formula: str
        """

        # generate bulk structure
        chemical_formula = formula
        v_per_atom = get_v_per_atom(formula)
        atoms = get_ase_atom_from_formula(chemical_formula, v_per_atom)
        atoms = sort(atoms)
        atoms.write('POSCAR')

        # get POTCAR
        os.system('getpotential_auto_zj.sh 6')

        # get INCAR
        with open('INCAR', 'w') as f:
            f.write(self.hp_config['INCAR_static'])
        os.system('modify_MAGMOM_auto.sh POSCAR')
        modify_INCAR('ISIF', '3')

        # get KPOINTS
        with open('KPOINTS', 'w') as f:
            f.write(self.hp_config['KPOINTS'])

        # run vasp
        run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
        print('Bulk static optimization done.')

    def surf_opt(self, bulk_structural_file='CONTCAR_bulk_opt'):
        """Structural optimization of the surface slab with VASP.

        :param bulk_structural_file: optimized bulk structure, defaults to 'CONTCAR_bulk_opt'
        :type bulk_structural_file: str, optional
        """

        atoms_bulk = read(bulk_structural_file)
        atoms_bulk.positions += 1.3 # avoid PBC error
        atoms_bulk.wrap() # avoid PBC error

        # add vacuum space and fix bottom atoms
        len_z = atoms_bulk.cell.array[2][2]
        c     = FixAtoms(indices=np.where(atoms_bulk.positions[:,2] < len_z / 2)[0])
        atoms_bulk.set_constraint(c)
        add_vacuum(atoms_bulk, 10.0)
        atoms_bulk.write('POSCAR') # Write POSCAR of the surface model.

        # get POTCAR
        os.system('getpotential_auto_zj.sh 6')

        # get INCAR
        with open('INCAR', 'w') as f:
            f.write(self.hp_config['INCAR_static'])
        os.system('modify_MAGMOM_auto.sh POSCAR')
        modify_INCAR('ISIF', '2')

        # get KPOINTS
        with open('KPOINTS', 'w') as f:
            f.write(self.hp_config['KPOINTS'])

        # run vasp
        run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
        print('Surface static optimization done.')

    def ads_opt(self, structural_file='CONTCAR_surf_opt', random_samples=5):
        """Structural optimization of the adsorption with VASP.

        :param structural_file: Structural file name of optimized clean surface, defaults to 'CONTCAR_surf_opt'
        :type structural_file: str, optional
        :param random_samples: On one surface, many surface sites can be detected, this number controls how many individual calculations will be performed on this surface, defaults to 5
        :type random_samples: int, optional

        .. Note::
            ``random_samples`` cannot be larger than the number of detected surface sites.
        """

        # generate structures
        for ads in self.hp_config['adsorbates']:
            # generate adsorption configurations
            adder     = AddAtoms(structural_file,
                                 species=ads,
                                 sites=self.hp_config['sites'],
                                 dist_from_surf=self.hp_config['dist_from_surf'],
                                 num_atomic_layer_along_Z=6)
            all_sites = adder.write_file_with_adsorption_sites(adsorbate_poscar,
                                                               calculation_index=self.hp_config['calculation_index'])

            if not os.path.exists('POSCARs'):
                os.mkdir('POSCARs')
            for i in range(all_sites):
                os.rename(f'POSCAR_{self.hp_config["calculation_index"]}_{i}',
                          os.path.join('POSCARs', f'POSCAR_{self.hp_config["calculation_index"]}_{i}'))

            for i in np.random.choice(range(all_sites), random_samples, replace=False):
                if not os.path.exists(str(i)):
                    os.mkdir(str(i))
                shutil.copyfile(os.path.join('POSCARs', f'POSCAR_{self.hp_config["calculation_index"]}_{i}'),
                               os.path.join(str(i), 'POSCAR'))

                parent_dir = os.getcwd()
                os.chdir(str(i))
                # get POTCAR
                os.system('getpotential_auto_zj.sh 6')

                # get INCAR
                with open('INCAR', 'w') as f:
                    f.write(self.hp_config['INCAR_static'])
                os.system('modify_MAGMOM_auto.sh POSCAR')
                modify_INCAR('ISIF', '2')

                # get KPOINTS
                with open('KPOINTS', 'w') as f:
                    f.write(self.hp_config['KPOINTS'])

                # run vasp
                run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
                print(f'Adsorption static calculation for {ads}_{i} done.')
                os.chdir(parent_dir)

    def bulk_aimd(self, formula):
        """AIMD simulation for a bulk structure of given chemical formula.

        :param formula: The given chemical formula.
        :type formula: str
        """

        # generate bulk structure
        chemical_formula = formula
        v_per_atom = get_v_per_atom(formula)
        atoms = get_ase_atom_from_formula(chemical_formula, v_per_atom)
        atoms = sort(atoms)
        atoms.write('POSCAR')

        # get POTCAR
        os.system('getpotential_auto_zj.sh 6')

        # get INCAR
        with open('INCAR', 'w') as f:
            f.write(self.hp_config['INCAR_aimd'])
        os.system('modify_MAGMOM_auto.sh POSCAR')
        modify_INCAR('ISIF', '3')
        modify_INCAR('NSW', '100')

        # get KPOINTS
        with open('KPOINTS', 'w') as f:
            f.write(self.hp_config['KPOINTS'])

        # run vasp
        run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
        print('Bulk AIMD simulation done.')

    def surface_aimd(self, bulk_structural_file='CONTCAR_bulk_opt'):
        """AIMD simulation for the clean surface.

        :param bulk_structural_file: File name of the bulk structure, defaults to 'CONTCAR_bulk_opt'
        :type bulk_structural_file: str, optional
        """

        atoms_bulk = read(bulk_structural_file)
        atoms_bulk.positions += 1.3 # avoid PBC error
        atoms_bulk.wrap() # avoid PBC error

        # add vacuum space and fix bottom atoms
        len_z = atoms_bulk.cell.array[2][2]
        c     = FixAtoms(indices=np.where(atoms_bulk.positions[:,2] < len_z / 2)[0])
        atoms_bulk.set_constraint(c)
        add_vacuum(atoms_bulk, 10.0)
        atoms_bulk.write('POSCAR') # Write POSCAR of the surface model.

        # get POTCAR
        os.system('getpotential_auto_zj.sh 6')

        # get INCAR
        with open('INCAR', 'w') as f:
            f.write(self.hp_config['INCAR_aimd'])
        os.system('modify_MAGMOM_auto.sh POSCAR')
        modify_INCAR('ISIF', '2')
        modify_INCAR('NSW', '100')

        # get KPOINTS
        with open('KPOINTS', 'w') as f:
            f.write(self.hp_config['KPOINTS'])

        # run vasp
        run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
        print('Surface AIMD simulation done.')

    def ads_aimd(self, structural_file='CONTCAR_surf_opt', random_samples=2):
        """AIMD simulation for the adsorption.

        :param structural_file: File name of the clean surface, defaults to 'CONTCAR_surf_opt'
        :type structural_file: str, optional
        :param random_samples: Randomly select surface sites for the simulation, defaults to 2
        :type random_samples: int, optional

        .. Note::
            ``random_samples`` cannot be larger than the number of detected surface sites.
        """

        # generate structures
        for ads in self.hp_config['adsorbates']:
            # generate adsorption configurations
            adder     = AddAtoms(structural_file,
                                 species=ads,
                                 sites=self.hp_config['sites'],
                                 dist_from_surf=self.hp_config['dist_from_surf'],
                                 num_atomic_layer_along_Z=6)
            all_sites = adder.write_file_with_adsorption_sites(adsorbate_poscar,
                                                               calculation_index=self.hp_config['calculation_index'])

            if not os.path.exists('POSCARs'):
                os.mkdir('POSCARs')
            for i in range(all_sites):
                os.rename(f'POSCAR_{self.hp_config["calculation_index"]}_{i}',
                          os.path.join('POSCARs', f'POSCAR_{self.hp_config["calculation_index"]}_{i}'))

            for i in np.random.choice(range(all_sites), random_samples, replace=False):
                if not os.path.exists(str(i)):
                    os.mkdir(str(i))
                shutil.copyfile(os.path.join('POSCARs', f'POSCAR_{self.hp_config["calculation_index"]}_{i}'),
                               os.path.join(str(i), 'POSCAR'))

                parent_dir = os.getcwd()
                os.chdir(str(i))
                # get POTCAR
                os.system('getpotential_auto_zj.sh 6')

                # get INCAR
                with open('INCAR', 'w') as f:
                    f.write(self.hp_config['INCAR_aimd'])
                os.system('modify_MAGMOM_auto.sh POSCAR')
                modify_INCAR('ISIF', '2')
                modify_INCAR('NSW', '100')

                # get KPOINTS
                with open('KPOINTS', 'w') as f:
                    f.write(self.hp_config['KPOINTS'])

                # run vasp
                run_vasp(vasp_bash_path=self.hp_config['vasp_bash_path'])
                print(f'Adsorption AIMD simulation for {ads}_{i} done.')
                os.chdir(parent_dir)

    def run(self, formula, **kwargs):
        """

        :param formula: Chemical formula.
        :type formula: str
        :param **kwargs: Configurations to control the process
        :type **kwargs: dict

        """

        self.hp_config = {**self.hp_config, **config_parser(kwargs)}

        # bulk static calculation
        if self.hp_config['include_bulk_static']:
            if not os.path.exists('bulk_static'):
                os.mkdir('bulk_static')
            os.chdir('bulk_static')
            self.bulk_opt(formula)
            os.chdir(self.root_dir)

        # surface static calculation
        if self.hp_config['include_surface_static']:
            if not os.path.exists('surface_static'):
                os.mkdir('surface_static')
            shutil.copyfile(os.path.join('bulk_static', 'CONTCAR'),
                            os.path.join('surface_static', 'CONTCAR_bulk_opt'))
            os.chdir('surface_static')
            self.surf_opt(bulk_structural_file='CONTCAR_bulk_opt')
            os.chdir(self.root_dir)

        # adsorption static calculation
        if self.hp_config['include_adsorption_static']:
            if not os.path.exists('adsorption_static'):
                os.mkdir('adsorption_static')
            shutil.copyfile(os.path.join('surface_static', 'CONTCAR'),
                            os.path.join('adsorption_static', 'CONTCAR_surf_opt'))
            os.chdir('adsorption_static')
            self.ads_opt(structural_file='CONTCAR_surf_opt', random_samples=self.hp_config['random_samples'])
            os.chdir(self.root_dir)

        # bulk aimd
        if self.hp_config['include_bulk_aimd']:
            if not os.path.exists('bulk_aimd'):
                os.mkdir('bulk_aimd')
            os.chdir('bulk_aimd')
            self.bulk_aimd(formula)
            os.chdir(self.root_dir)

        # surface aimd calculation
        if self.hp_config['include_surface_aimd']:
            if not os.path.exists('surface_aimd'):
                os.mkdir('surface_aimd')
            shutil.copyfile(os.path.join('bulk_static', 'CONTCAR'),
                            os.path.join('surface_aimd', 'CONTCAR_bulk_opt'))
            os.chdir('surface_aimd')
            self.surface_aimd(bulk_structural_file='CONTCAR_bulk_opt')
            os.chdir(self.root_dir)

        # adsorption static calculation
        if self.hp_config['include_adsorption_aimd']:
            if not os.path.exists('adsorption_aimd'):
                os.mkdir('adsorption_aimd')
            shutil.copyfile(os.path.join('surface_static', 'CONTCAR'),
                            os.path.join('adsorption_aimd', 'CONTCAR_surf_opt'))
            os.chdir('adsorption_aimd')
            self.ads_aimd(structural_file='CONTCAR_surf_opt', random_samples=self.hp_config['random_samples'])
            os.chdir(self.root_dir)

if __name__ == '__main__':
    HT = HtDftAds(calculation_index=0)
    HT.run(sys.argv[1]) # debug only



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

from .build_dataset import CrystalGraph, ReadGraphs, ExtractVaspFiles, BuildDatabase, concat_graphs, select_graphs_random
from .atomic_feature import get_atomic_feature_onehot
from .load_dataset import LoadDataset, Collater

from ..lib.model_lib import load_graph_build_method



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 15:30:20 2023

@author: ZHANG Jun
"""

import os
import json
import torch
import torch.nn as nn

from dgl.ops import edge_softmax
from dgl import function as fn
from dgl.data.utils import load_graphs

from .layer import Layer

class PotentialModel(nn.Module):
    """A GAT model with multiple gat layers for predicting atomic energies, forces, and stress tensors.


    .. Note:: You can also use this model to train and predict atom and bond related properties. You need to store the labels on graph edges if you want to do so. This model has multiple attention heads.


    .. Important::

        The first value of ``gat_node_dim_list`` is the depth of atomic representation.

        The first value of ``energy_readout_node_list``, ``force_readout_node_list``, ``stress_readout_node_list`` is the input dimension and equals to last value of `gat_node_list * num_heads`.

        The last values of ``energy_readout_node_list``, ``force_readout_node_list``, ``stress_readout_node_list`` are ``1``, ``3``, and ``6``, respectively.


    :param gat_node_dim_list: A list of node dimensions of the AGAT ``Layer``s.
    :type gat_node_dim_list: list
    :param energy_readout_node_list: A list of node dimensions of the energy readout layers.
    :type energy_readout_node_list: list
    :param force_readout_node_list: A list of node dimensions of the force readout layers.
    :type force_readout_node_list: list
    :param stress_readout_node_list: A list of node dimensions of the stress readout layers.
    :type stress_readout_node_list: list
    :param head_list: A list of attention head names, defaults to ['div']
    :type head_list: list, optional
    :param bias: Add bias or not to the neural networks., defaults to True
    :type bias: TYPE, bool
    :param negative_slope: This specifies the negative slope of the LeakyReLU (see https://pytorch.org/docs/stable/generated/torch.nn.LeakyReLU.html) activation function., defaults to 0.2
    :type negative_slope: float, optional
    :param device: Device to train the model. Use GPU cards to accerelate training., defaults to 'cuda'
    :type device: str, optional
    :param tail_readout_no_act: The tail ``tail_readout_no_act`` layers will have no activation functions. The first, second, and third elements are for energy, force, and stress readout layers, respectively., defaults to [3,3,3]
    :type tail_readout_no_act: list, optional
    :return: An AGAT model
    :rtype: agat.model.PotentialModel

    """
    def __init__(self,
                 gat_node_dim_list,
                 energy_readout_node_list,
                 force_readout_node_list,
                 stress_readout_node_list,
                 head_list=['div'],
                 bias=True,
                 negative_slope=0.2,
                 device = 'cuda',
                 tail_readout_no_act=[3,3,3]):
        super(PotentialModel, self).__init__()

        # args
        self.gat_node_dim_list = gat_node_dim_list
        self.energy_readout_node_list = energy_readout_node_list
        self.force_readout_node_list = force_readout_node_list
        self.stress_readout_node_list = stress_readout_node_list
        self.head_list = head_list
        self.bias = bias
        self.device = device
        self.negative_slope = negative_slope
        self.tail_readout_no_act = tail_readout_no_act

        self.num_gat_layers = len(self.gat_node_dim_list)-1
        self.num_energy_readout_layers = len(self.energy_readout_node_list)-1
        self.num_force_readout_layers = len(self.force_readout_node_list)-1
        self.num_stress_readout_layers = len(self.stress_readout_node_list)-1
        self.num_heads = len(self.head_list)

        self.__gat_real_node_dim_list = [x*self.num_heads for x in self.gat_node_dim_list[1:]]
        self.__gat_real_node_dim_list.insert(0,self.gat_node_dim_list[0])
        # self.energy_readout_node_list.insert(0, self.__gat_real_node_dim_list[-1]) # calculate and insert input dimension.
        # self.force_readout_node_list.insert(0, self.__gat_real_node_dim_list[-1])
        # self.stress_readout_node_list.insert(0, self.__gat_real_node_dim_list[-1])

        # register layers and parameters.
        self.gat_layers = nn.ModuleList([])
        self.energy_readout_layers = nn.ModuleList([])
        self.force_readout_layers = nn.ModuleList([])
        self.stress_readout_layers = nn.ModuleList([])

        for l in range(self.num_gat_layers):
            self.gat_layers.append(Layer(self.__gat_real_node_dim_list[l],
                                            self.gat_node_dim_list[l+1],
                                            self.num_heads,
                                            device=self.device,
                                            bias=self.bias,
                                            negative_slope=self.negative_slope))

        # energy readout layer
        for l in range(self.num_energy_readout_layers-self.tail_readout_no_act[0]):
            self.energy_readout_layers.append(nn.Linear(self.energy_readout_node_list[l],
                                                         self.energy_readout_node_list[l+1],
                                                         self.bias, self.device))
            self.energy_readout_layers.append(nn.LeakyReLU(negative_slope=self.negative_slope))
        for l in range(self.tail_readout_no_act[0]):
            self.energy_readout_layers.append(nn.Linear(self.energy_readout_node_list[l-self.tail_readout_no_act[0]-1],
                                                         self.energy_readout_node_list[l-self.tail_readout_no_act[0]],
                                                         self.bias, self.device))
        # force readout layer
        for l in range(self.num_force_readout_layers-self.tail_readout_no_act[1]):
            self.force_readout_layers.append(nn.Linear(self.force_readout_node_list[l],
                                                        self.force_readout_node_list[l+1],
                                                        self.bias, self.device))
            self.force_readout_layers.append(nn.LeakyReLU(negative_slope=self.negative_slope))
        for l in range(self.tail_readout_no_act[1]):
            self.force_readout_layers.append(nn.Linear(self.force_readout_node_list[l-self.tail_readout_no_act[1]-1],
                                                        self.force_readout_node_list[l-self.tail_readout_no_act[1]],
                                                        self.bias, self.device))
        # stress readout layer
        for l in range(self.num_stress_readout_layers-self.tail_readout_no_act[2]):
            self.stress_readout_layers.append(nn.Linear(self.stress_readout_node_list[l],
                                                         self.stress_readout_node_list[l+1],
                                                         self.bias, self.device))
            self.stress_readout_layers.append(nn.LeakyReLU(negative_slope=self.negative_slope))
        for l in range(self.tail_readout_no_act[2]):
            self.stress_readout_layers.append(nn.Linear(self.stress_readout_node_list[l-self.tail_readout_no_act[2]-1],
                                                         self.stress_readout_node_list[l-self.tail_readout_no_act[2]],
                                                         self.bias, self.device))

        self.__real_num_energy_readout_layers = len(self.energy_readout_layers)
        self.__real_num_force_readout_layers = len(self.force_readout_layers)
        self.__real_num_stress_readout_layers = len(self.stress_readout_layers)

        # attention heads
        self.head_fn = {'mul' : self.mul,
                        'div' : self.div,
                        'free': self.free}

    def mul(self, TorchTensor):
        """Multiply head.

        :param TorchTensor: Input tensor
        :type TorchTensor: torch.tensor
        :return: Ouput tensor
        :rtype: torch.tensor

        """

        return TorchTensor

    def div(self, TorchTensor):
        """Division head.

        :param TorchTensor: Input tensor
        :type TorchTensor: torch.tensor
        :return: Ouput tensor
        :rtype: torch.tensor

        """
        return 1/TorchTensor

    def free(self, TorchTensor):
        """Free head.

        :param TorchTensor: Input tensor
        :type TorchTensor: torch.tensor
        :return: Ouput tensor all ones
        :rtype: torch.tensor

        """
        return torch.ones(TorchTensor.size(), device=self.device)

    def get_head_mechanism(self, fn_list, TorchTensor):
        """Get attention heads

        :param fn_list: A list of head mechanisms. For example: ['mul', 'div', 'free']
        :type fn_list: list
        :param TorchTensor: A PyTorch tensor
        :type TorchTensor: torch.Tensor
        :return: A new tensor after the transformation.
        :rtype: torch.Tensor

        """
        TorchTensor_list = []
        for func in fn_list:
            TorchTensor_list.append(self.head_fn[func](TorchTensor))
        return torch.cat(TorchTensor_list, 1)

    def forward(self, graph):
        """The ``forward`` function of PotentialModel model.

        :param graph: ``DGL.Graph``
        :type graph: ``DGL.Graph``
        :return:
            - energy: atomic energy
            - force: atomic force
            - stress: cell stress tensor

        :rtype: tuple of torch.tensors

        """
        with graph.local_scope():
            h    = graph.ndata['h']                                    # shape: (number of nodes, dimension of one-hot code representation)
            dist = torch.reshape(graph.edata['dist'], (-1, 1, 1))      # shape: (number of edges, 1, 1)
            dist = torch.where(dist < 0.01, 0.01, dist)                  # This will creat a new `dist` variable, insted of modifying the original memory.
            dist = self.get_head_mechanism(self.head_list, dist) # shape of dist: (number of edges, number of heads, 1)

            for l in range(self.num_gat_layers):
                h = self.gat_layers[l](h, dist, graph)                 # shape of h: (number of nodes, number of heads * num_out)

            # predict energy
            energy = h
            for l in range(self.__real_num_energy_readout_layers):
                energy = self.energy_readout_layers[l](energy)
            batch_nodes = graph.batch_num_nodes().tolist()
            energy = torch.split(energy, batch_nodes)
            energy = torch.stack([torch.mean(e) for e in energy])

            # Predict force
            graph.ndata['node_force'] = h
            graph.apply_edges(fn.u_add_v('node_force', 'node_force', 'force_score'))   # shape of score: (number of edges, ***, 1)
            force_score = torch.reshape(graph.edata['force_score'],(-1, self.num_heads, self.gat_node_dim_list[-1])) / dist
            force_score = torch.reshape(force_score, (-1, self.num_heads * self.gat_node_dim_list[-1]))

            stress_score = torch.clone(force_score)

            for l in range(self.__real_num_force_readout_layers):
                force_score = self.force_readout_layers[l](force_score)
            graph.edata['force_score_vector'] = force_score * graph.edata['direction']      # shape (number of edges, 1)
            graph.update_all(fn.copy_e('force_score_vector', 'm'), fn.sum('m', 'force_pred'))        # shape of graph.ndata['force_pred']: (number of nodes, 3)
            force = graph.ndata['force_pred']

            # Predict stress
            for l in range(self.__real_num_stress_readout_layers):
                stress_score = self.stress_readout_layers[l](stress_score)
            graph.edata['stress_score_vector'] = stress_score * torch.cat((graph.edata['direction'],graph.edata['direction']), dim=1)      # shape (number of edges, 1)
            graph.update_all(fn.copy_e('stress_score_vector', 'm'), fn.sum('m', 'stress_pred'))        # shape of graph.ndata['force_pred']: (number of nodes, 3)
            # stress = torch.sum(graph.ndata['stress_pred'], dim=0)
            stress = torch.split(graph.ndata['stress_pred'], batch_nodes)
            stress = torch.stack([torch.sum(s, dim=0) for s in stress])
            return energy, force, stress

class CrystalPropertyModel(nn.Module):
    pass

class AtomicPropertyModel(nn.Module):
    pass

class AtomicVectorModel(nn.Module):
    pass

if __name__ == '__main__':
    import dgl
    g_list, l_list = dgl.load_graphs('all_graphs.bin')
    graph = g_list[1] #.to('cuda')
    graph = g_list[1].to('cuda')
    feat = graph.ndata['h']
    dist = graph.edata['dist']

    PM = PotentialModel([6,30,20,10],
                        [10,10,10,5,1],
                        [10,10,10,5,3],
                        [10,10,10,5,6],
                        head_list=['div', 'mul'],
                        bias=True,
                        negative_slope=0.2,
                        device = 'cuda',
                        tail_readout_no_act=[2,2,2]
                        )

    energy, force, stress = PM.forward(graph)

    model = PM



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

__version__ = '8.0.5'

import os
# os.environ['DGLBACKEND']="pytorch"

# import important object.
from .data.build_dataset import BuildDatabase
from .model.fit import Fit
from .app.cata.high_throughput_predict import HtAds

del os



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

def main(formula):
    from agat.data.build_dataset import BuildDatabase
    from agat.model.fit import Fit
    from agat.app.cata import high_throughput_predict
    from agat.app.cata.high_throughput_predict import HtAds

    ad = BuildDatabase(mode_of_NN='ase_dist', num_of_cores=16)
    ad.build()

    f = Fit()
    f.fit()

    HA = HtAds()
    HA.run(formula, device='cuda')



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 26 23:07:24 2023

@author: ZHANG Jun
"""

'''

.. py:data:: config

    Configurations for database construction, training process, and prediction behaviors.

'''

import os
import torch.nn as nn

default_elements = ['Ac', 'Ag', 'Al', 'Am', 'Ar', 'As', 'At', 'Au', 'B',  'Ba',
                    'Be', 'Bh', 'Bi', 'Bk', 'Br', 'C',  'Ca', 'Cd', 'Ce', 'Cf',
                    'Cl', 'Cm', 'Cn', 'Co', 'Cr', 'Cs', 'Cu', 'Db', 'Ds', 'Dy',
                    'Er', 'Es', 'Eu', 'F',  'Fe', 'Fl', 'Fm', 'Fr', 'Ga', 'Gd',
                    'Ge', 'H',  'He', 'Hf', 'Hg', 'Ho', 'Hs', 'I',  'In', 'Ir',
                    'K',  'Kr', 'La', 'Li', 'Lr', 'Lu', 'Lv', 'Mc', 'Md', 'Mg',
                    'Mn', 'Mo', 'Mt', 'N',  'Na', 'Nb', 'Nd', 'Ne', 'Nh', 'Ni',
                    'No', 'Np', 'O',  'Og', 'Os', 'P',  'Pa', 'Pb', 'Pd', 'Pm',
                    'Po', 'Pr', 'Pt', 'Pu', 'Ra', 'Rb', 'Re', 'Rf', 'Rg', 'Rh',
                    'Rn', 'Ru', 'S',  'Sb', 'Sc', 'Se', 'Sg', 'Si', 'Sm', 'Sn',
                    'Sr', 'Ta', 'Tb', 'Tc', 'Te', 'Th', 'Ti', 'Tl', 'Tm', 'Ts',
                    'U',  'V',  'W',  'Xe', 'Y',  'Yb', 'Zn', 'Zr']

default_build_properties = {'energy': True,
                         'forces': True,
                         'cell': True,
                         'cart_coords': True,
                         'frac_coords': True,
                         'constraints': True,
                         'stress': True,
                         'distance': True,
                         'direction': True,
                         'path': False}

default_data_config =  {
    'species': default_elements,
    'path_file': 'paths.log', # A file of absolute paths where OUTCAR and XDATCAR files exist.
    'build_properties': default_build_properties, # Properties needed to be built into graph.
    'topology_only': False,
    'dataset_path': 'dataset', # Path where the collected data to save.
    'mode_of_NN': 'ase_natural_cutoffs', # How to identify connections between atoms. 'ase_natural_cutoffs', 'pymatgen_dist', 'ase_dist', 'voronoi'. Note that pymatgen is much faster than ase.
    'cutoff': 5.0, # Cutoff distance to identify connections between atoms. Deprecated if ``mode_of_NN`` is ``'ase_natural_cutoffs'``
    'load_from_binary': False, # Read graphs from binary graphs that are constructed before. If this variable is ``True``, these above variables will be depressed.
    'num_of_cores': 8,
    'super_cell': False,
    'has_adsorbate': False,
    'keep_readable_structural_files': False,
    'mask_similar_frames': False,
    'mask_reversed_magnetic_moments': False, # or -0.5 # Frames with atomic magnetic moments lower than this value will be masked.
    'energy_stride': 0.05,
    'scale_prop': False
             }

FIX_VALUE = [1,3,6]

default_train_config = {
    'verbose': 1, # `0`: no train and validation output; `1`: Validation and test output; `2`: train, validation, and test output.
    'dataset_path': os.path.join('dataset', 'all_graphs.bin'),
    'model_save_dir': 'agat_model',
    'epochs': 1000,
    'output_files': 'out_file',
    'device': 'cuda:0',
    'validation_size': 0.15,
    'test_size': 0.15,
    'early_stop': True,
    'stop_patience': 300,
    'head_list': ['mul', 'div', 'free'],
    'gat_node_dim_list': [len(default_elements), 100, 100, 100],
    'energy_readout_node_list': [300, 100, 50, 30, 10, 3, FIX_VALUE[0]], # the first value should be: len(head_list)*gat_node_dim_list[-1]
    'force_readout_node_list': [300, 100, 50, 30, 10, FIX_VALUE[1]], # the first value should be: len(head_list)*gat_node_dim_list[-1]
    'stress_readout_node_list': [300, 100, 50, 30, 10, FIX_VALUE[2]], # the first value should be: len(head_list)*gat_node_dim_list[-1]
    'bias': True,
    'negative_slope': 0.2,
    'criterion': nn.MSELoss(),
    'a': 1.0,
    'b': 1.0,
    'c': 0.0,
    # 'optimizer': 'adam',
    'learning_rate': 0.0001,
    'weight_decay': 0.0, # weight decay (L2 penalty)
    'batch_size': 64,
    'val_batch_size': 400,
    'transfer_learning': False,
    'trainable_layers': -4,
    'mask_fixed': False,
    'tail_readout_no_act': [3,3,3],
    # 'adsorbate': False, #  or not when building graphs.
    'adsorbate_coeff': 20.0, # indentify and specify the importance of adsorbate atoms with respective to surface atoms. zero for equal importance.
    'transfer_learning': False}

default_ase_calculator_config = {'fmax'             : 0.1,
                                 'steps'            : 200,
                                 'maxstep'          : 0.05,
                                 'restart'          : None,
                                 'restart_steps'    : 0,
                                 'perturb_steps'    : 0,
                                 'perturb_amplitude': 0.05,
                                 'out'              : None}


default_high_throughput_config = {
        'model_save_dir': 'agat_model',
        'opt_config': default_ase_calculator_config,
        'calculation_index'    : '0', # sys.argv[1],
        'fix_all_surface_atom' : False,
        'remove_bottom_atoms'  : False,
        'save_trajectory'      : False,
        'partial_fix_adsorbate': True,
        'adsorbates'           : ['H'],
        'sites'                : ['ontop'],
        'dist_from_surf'       : 1.7,
        'using_template_bulk_structure': False,
        'graph_build_scheme_dir': os.path.join('dataset'),
        'device': 'cuda' # in our test results, the A6000 is about * times faster than EPYC 7763.
        }

default_predict_config = {}

default_active_learning_config = {}

default_hp_dft_config = {'INCAR_static': '''
SYSTEM = ML

Start parameter for this Run:
  ISTART = 0
  ICHARG = 2
  INIWAV = 1

Electronic Relaxation:
  ENCUT = 500
  PREC = Accurate
  ALGO = Fast
  NELM = 300
  NELMIN = 4
  EDIFF = 1E-06
  GGA = PE
  LREAL = A

Ionic Relaxation:
  EDIFFG = -0.05
  NSW = 300
  IBRION = 2
  ISIF = 3
  POTIM = 0.5

DOS related values:
  SIGMA = 0.1
  ISMEAR = 1

Spin polarized:
  ISPIN = 2
  MAGMOM = placeholder

File writing
  LWAVE = .FALSE.
  LCHARG = .FALSE.

Calculation of DOS
  NPAR = 8
  LORBIT = 11
  NCORE = 1
  IVDW = 11

''',

'INCAR_aimd': '''
SYSTEM = ML

Start parameter for this Run:
  ISTART = 0
  ICHARG = 2
  INIWAV = 1

Electronic Relaxation:
  ENCUT = 500
  PREC = Accurate
  ALGO = Fast
  NELM = 300
  NELMIN = 4
  EDIFF = 1E-06
  GGA = PE
  LREAL = A

Ionic Relaxation:
  NSW = 100
  IBRION = 0
  ISIF = 3
  POTIM = 2

DOS related values:
  SIGMA = 0.1
  ISMEAR = 1

Spin polarized:
  ISPIN = 2
  MAGMOM = placeholder

File writing
  LWAVE = .FALSE.
  LCHARG = .FALSE.

Calculation of DOS
  NPAR = 8
  LORBIT = 11
  IVDW = 11


  MDALGO = 3
  SMASS = 0
  TEBEG = 300
  TEEND = 300

''',

  "KPOINTS": '''Automatic mesh
0
Gamma
1 1 1
0.0 0.0 0.0

''',

        'calculation_index'    : '0', # sys.argv[1],
        'adsorbates'           : ['H'],
        'sites'                : ['bridge'],
        'dist_from_surf'       : 1.7,
        'include_bulk_static': True, # This should be true for other calculations.
        'include_surface_static': True, # This should be true for other calculations.
        'include_adsorption_static': True,
        'include_bulk_aimd': True,
        'include_surface_aimd': True,
        'include_adsorption_aimd': True,
        'random_samples': 1, # number of surfaces
        'vasp_bash_path': 'vasp_run.sh'
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Mon Apr  4 18:12:17 2022

@author: ZHANG Jun
"""

from ase.io import read, write
import sys

def increase_vacuum(fname):
    """

    :param fname: Structural file name.
    :type fname: str
    :return: A new file with thicker vacuum.
    :rtype: A new file

    """

    # fname = sys.argv[1]

    poscar = read(fname)
    # frac   = poscar.get_scaled_positions()
    cart   = poscar.get_positions()
    cell   = poscar.cell.array
    cell[2][2] += 5

    poscar.set_cell(cell)
    poscar.set_positions(cart)

    write(f'{fname}_more_vacuum', poscar)



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 30 23:54:56 2023

@author: 18326
"""

import torch
from torch.utils.data import Dataset
import dgl
from dgl.data.utils import load_graphs

class LoadDataset(Dataset):
    """Load the binary graphs.

    Example::

        import os
        from agat.data import LoadDataset
        dataset=LoadDataset(os.path.join('dataset', 'all_graphs.bin'))

        # you can index or slice the dataset.
        g0, props0 = dataset[0]
        g_batch, props = dataset[0:100] # the g_batch is a batch collection of graphs. See https://docs.dgl.ai/en/1.1.x/generated/dgl.batch.html


    :param dataset_path: A paths leads to the binary DGL graph file.
    :type dataset_path: str
    :return: a graph dataset.
    :rtype: list

    """
    def __init__(self, dataset_path):
        super(LoadDataset, self).__init__()
        self.dataset_path = dataset_path
        self.graph_list, self.props = load_graphs(self.dataset_path) # `props`: properties.

    def __getitem__(self, index):
        """Index or slice the dataset.

        :param index: list index or slice
        :type index: int/slice
        :return: graph or graph batch
        :rtype: dgl graph
        :return: props. Graph labels
        :rtype: A dict of torch.tensor

        """

        if isinstance(index, slice):
            graph_list = self.graph_list[index]
            graph = dgl.batch(graph_list)
        else:
            graph = self.graph_list[index]
        props = {k:v[index] for k,v in self.props.items()}
        return graph, props

    def __len__(self):
        """Get the length of the dataset.

        :return: the length of the dataset
        :rtype: int

        """

        return len(self.graph_list)

# class LoadDataset(Dataset):
#     def __init__(self, dataset_path):
#         super(LoadDataset, self).__init__()
#         self.dataset_path = dataset_path
#         self.graph_list, self.props = load_graphs(self.dataset_path) # `props`: properties.

#     def __getitem__(self, index):
#         graph = self.graph_list[index]
#         props = {k:v[index] for k,v in self.props.items()}
#         return graph, props

#     def __len__(self):
#         return len(self.graph_list)

class Collater(object):
    """The collate function used in torch.utils.data.DataLoader: https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader

    The collate function determines how to merge the batch data.

    Example::

        import os
        from agat.data import LoadDataset, Collater
        from torch.utils.data import DataLoader

        dataset=LoadDataset(os.path.join('dataset', 'all_graphs.bin'))
        collate_fn = Collater(device='cuda')
        data_loader = DataLoader(dataset, batch_size=64, shuffle=True, collate_fn=collate_fn)

    :param device: device to store the merged data, defaults to 'cuda'
    :type device: str, optional

    """
    def __init__(self, device='cuda'):
        self.device = device

    def __call__(self, data):
        """Collate the data into batches.

        :param data: the output of :py:class:`LoadDataset`
        :type data: tuple
        :return: dgl batch graphs. See https://docs.dgl.ai/en/1.1.x/generated/dgl.batch.html
        :rtype: DGLGraph
        :return: Graph labels
        :rtype: A dict of torch.tensor

        """

        # print(data)
        graph_list = [x[0] for x in data]
        graph = dgl.batch(graph_list)
        props = [x[1] for x in data]
        props = {k:torch.stack([x[k] for x in props]) for k in props[0].keys()}
        props = {k:v.to(self.device) for k,v in props.items()}
        return graph.to(self.device), props

if __name__ == '__main__':
    dataset = LoadDataset('all_graphs.bin')
    g, p = dataset[10]
    from torch.utils.data import DataLoader
    collate_fn = Collater(device='cpu')
    import time
    start = time.time()
    loader = DataLoader(dataset, batch_size=64, shuffle=True, num_workers=4,
                        collate_fn=collate_fn)

    for i, sample in enumerate(loader):
        graph, props = sample
    print(time.time()-start)

    # test of num_workers larger than 0.
    start = time.time()
    for i, sample in enumerate(loader):
        graph, props = sample
        if i > 500:
            break
    dur1 = time.time()-start

    start = time.time()
    for i, sample in enumerate(loader):
        graph, props = sample
        time.sleep(0.01)
        if i > 500:
            break
    dur2 = time.time()-start
    print(dur2-5.0-dur1)




# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 21:59:30 2021

@author: ZHANG Jun
"""

import json
import sys
import os
import numpy as np

from ase.optimize import BFGS
from ase.io import read, write
from ase.build import add_vacuum, sort
from ase.constraints import FixAtoms

from ..app import AgatCalculator
from .generate_adsorption_sites import AddAtoms
from ...lib.file_lib import file_exit
from ...lib.model_lib import config_parser
from ...default_parameters import default_high_throughput_config
from ...lib.adsorbate_poscar import adsorbate_poscar
from ...lib.high_throughput_lib import get_v_per_atom, get_ase_atom_from_formula, get_ase_atom_from_formula_template

class HtAds(object):
    def __init__(self, **hp_config):
        self.hp_config = {**default_high_throughput_config, **config_parser(hp_config)}

    def geo_opt(self, atoms_with_calculator, **kwargs):
        calculator = atoms_with_calculator.get_calculator()
        atoms  = atoms_with_calculator.copy()
        atoms.set_calculator(calculator)

        config = {**self.hp_config['opt_config'], **kwargs}

        if isinstance(config["out"], type(None)):
            logfile    = '-'
            trajectory = None
        else:
            logfile    = f'{config["out"]}.log'
            trajectory = f'{config["out"]}.traj'
        force_opt, energy_opt, atoms_list = [], [], []
        for i in range(config["perturb_steps"]+1):
            dyn = BFGS(atoms,
                       logfile=logfile,
                       trajectory=trajectory,
                       restart=config["restart"],
                       maxstep=config["maxstep"])

            return_code  = dyn.run(fmax=config["fmax"], steps=config["steps"])
            restart_step = 1
            while not return_code and restart_step < config["restart_steps"]:
                restart_step += 1
                maxstep_tmp = config["maxstep"]/2**restart_step
                dyn = BFGS(atoms,
                           logfile=logfile,
                           trajectory=trajectory,
                           restart=config["restart"],
                           maxstep=maxstep_tmp)
                return_code  = dyn.run(fmax=config["fmax"], steps=config["steps"])
            force_opt.append(atoms.get_forces())
            energy_opt.append(atoms.get_potential_energy(apply_constraint=False))
            atoms_list.append(atoms.copy())
            if config["perturb_steps"] > 0:
                atoms = self.perturb_positions(atoms, amplitude=config["perturb_amplitude"])
        argmin = np.argmin(energy_opt)
        energy, force, atoms = energy_opt[argmin], force_opt[argmin], atoms_list[argmin]
        force_max = np.linalg.norm(force, axis=1).max()
        return energy, force, atoms, force_max

    def ads_calc(self, formula, calculator, **kwargs):
        hp_config =  {**self.hp_config, **kwargs}

        if hp_config['save_trajectory']:
            out_dir = f'{hp_config["calculation_index"]}_th_calculation'
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
        else:
            out_dir     = '.'
            outbasename = None

        # generate bulk structure
        chemical_formula = formula

        if self.hp_config['using_template_bulk_structure']:
            atoms = get_ase_atom_from_formula_template(chemical_formula,
                                                       get_v_per_atom(chemical_formula),
                                                       template_file='POSCAR_temp')
        else:
            atoms = get_ase_atom_from_formula(chemical_formula,
                                              v_per_atom=get_v_per_atom(chemical_formula))

        atoms.set_calculator(calculator)

        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'POSCAR_bulk.gat'), atoms, format='vasp')
            outbasename = os.path.join(out_dir, 'bulk_opt')

        hp_config['out'] = outbasename

        energy_bulk, force_bulk, atoms_bulk, force_max_bulk = self.geo_opt(atoms,
                                                                           **hp_config['opt_config'])
        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'CONTCAR_bulk.gat'), atoms_bulk)

        print('Bulk optimization done.')

        # add vacuum space and fix bottom atoms
        len_z = atoms_bulk.cell.array[2][2]
        atoms_bulk.positions += 1.3 # avoid PBC error
        atoms_bulk.wrap()
        c     = FixAtoms(indices=np.where(atoms_bulk.positions[:,2] < len_z / 2)[0])
        atoms_bulk.set_constraint(c)

        if hp_config['remove_bottom_atoms']:
            pop_list = np.where(atoms_bulk.positions[:,2] < 1.0)
            del atoms_bulk[pop_list]

        add_vacuum(atoms_bulk, 10.0)

        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'POSCAR_surface.gat'), atoms_bulk)
            outbasename = os.path.join(out_dir, 'surface_opt')

        # surface optimization
        atoms_bulk.set_calculator(calculator)
        hp_config['out'] = outbasename

        energy_surf, force_surf, atoms_surf, force_max_surf = self.geo_opt(atoms_bulk,
                                                                           **hp_config['opt_config'])
        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'CONTCAR_surface.gat'), atoms_surf)

        print('Surface optimization done.')

        if force_max_surf < hp_config['opt_config']['fmax']:
            if hp_config['fix_all_surface_atom']:
                c = FixAtoms(indices=[x for x in range(len(atoms_surf))])
                atoms_surf.set_constraint(c)
            write(f'POSCAR_surf_opt_{hp_config["calculation_index"]}.gat', sort(atoms_surf))

            # adsorbate_shift = {'bridge': 0.0, 'ontop': 0.35, 'hollow': -0.1}

            for ads in hp_config['adsorbates']:
                # generate adsorption configurations: OH adsorption
                adder     = AddAtoms(f'POSCAR_surf_opt_{hp_config["calculation_index"]}.gat',
                                     species=ads,
                                     sites=hp_config['sites'],
                                     dist_from_surf=hp_config['dist_from_surf'],
                                     num_atomic_layer_along_Z=6)
                all_sites = adder.write_file_with_adsorption_sites(adsorbate_poscar,
                                                                   calculation_index=hp_config['calculation_index'])

                # adsorption optimization
                ase_atoms = [read(f'POSCAR_{hp_config["calculation_index"]}_{x}') for x in range(all_sites)]
                [os.remove(f'POSCAR_{hp_config["calculation_index"]}_{x}') for x in range(all_sites)]
                # ase_atoms = ase_atoms[0:3] # !!!

                energy_ads_list, converge_stat = [], []
                for i, ads_atoms in enumerate(ase_atoms):
                    file_exit()
                    if hp_config['save_trajectory']:
                        write(os.path.join(out_dir, f'POSCAR_{ads}_ads_{hp_config["calculation_index"]}_{i}.gat'),
                              ads_atoms)
                        outbasename = os.path.join(out_dir, f'adsorption_{ads}_opt_{i}')

                    hp_config['out'] = outbasename
                    ads_atoms.set_calculator(calculator)
                    energy_ads, force_ads, atoms_ads, force_max_ads = self.geo_opt(ads_atoms, **hp_config['opt_config'])

                    if hp_config['save_trajectory']:
                        write(os.path.join(out_dir, f'CONTCAR_{ads}_ads_{hp_config["calculation_index"]}_{i}.gat'),
                              atoms_ads)

                    energy_ads_list.append(energy_ads)

                    if force_max_ads > hp_config['opt_config']['fmax']:
                        converge_stat.append(0.0)
                    else:
                        converge_stat.append(1.0)

                energy_surf_list = np.array([energy_surf] * len(energy_ads_list))
                energy_ads_list = np.array(energy_ads_list)

                out = np.vstack([energy_ads_list, energy_surf_list, converge_stat]).T
                np.savetxt(f'ads_surf_energy_{ads}_{hp_config["calculation_index"]}.txt',
                           out, fmt='%f')

    def run(self, formula, **kwargs):
        """

        :param formula: Input chemical formula
        :type formula: str
        :param **kwargs: DESCRIPTION
        :type **kwargs: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        # model save path
        model_save_dir = self.hp_config['model_save_dir']

        # instantiate a calculator
        calculator=AgatCalculator(model_save_dir,
                                 self.hp_config['graph_build_scheme_dir'],
                                 device=self.hp_config['device'])

        with open('high_throughput_config.json', 'w') as f:
            json.dump(self.hp_config, f, indent=4)

        self.ads_calc(formula, calculator, **kwargs)

if __name__ == '__main__':
    HA = HtAds()
    HA.run('NiCoFePdPt') # debug only



INCAR_TAG = '''
SYSTEM
ISTART
ICHARG
INIWAV
ENCUT
ENAUG
PREC
IALGO
NELM
NELMIN
NELMDL
EDIFF
NBANDS
GGA
VOSKOWN
LREAL
WEIMIN
EDIFFG
NSW
IBRION
ISIF
POTIM
IOPT
ISYM
SIGMA
ISMEAR
ISPIN
MAGMOM
LWAVE
LCHARG
RWIGS
NPAR
LORBIT
LDAU
LDAUTYPE
LDAUL
LDAUU
LDAUJ
LDAUPRINT
LMAXMIX
LASPH
IDIPOL
LDIPOL
LAECHG
LADDGRID
NGX
NGY
NGZ
NGXF
NGYF
NGZF
ICHAIN
IMAGES
SPRING
LCLIMB
DdR
DRotMax
DFNMin
DFNMax
NFREE
LUSE_VDW
Zab_vdW
AGGAC
AMIX
AMIX_MAG
BMIX
BMIX_MAG
ALGO
KPAR
NCORE
NEDOS
IVDW
LELF
MDALGO
'''.split()




# -*- coding: utf-8 -*-
"""
Created on Tue Sep  7 22:32:50 2021

@author: ZHANG Jun
"""

from io import StringIO
import numpy as np

from ase import Atoms, Atom
from ase.io import read, write
from ase.formula import Formula
from ase.build import add_adsorbate

class AddAtoms(object):
    def __init__(self, fname, sites='all', species='O',
                 num_atomic_layer_along_Z=3, dist_from_surf=2.2,
                 crystal_type='fcc'):
        self.fname                       = fname

        self._support_sites              = ['ontop', 'bridge', 'hollow', 'disigma']

        if isinstance(sites, str):
            self.sites = [sites]
        else:
            self.sites = sites
        if sites[0] == 'all':
            self.sites               = self._support_sites
        self.sites                   = list(np.reshape(self.sites, (1,-1))[0])

        assert set(self.sites) < set(self._support_sites + ['all']), 'Illegal site name.'

        self._structure                  = read(self.fname)
        # self.length_x                    = self._structure.cell.array[0][0]
        # self.length_y                    = self._structure.cell.array[1][1]
        self.length_z                    = self._structure.cell.array[2][2]
        self.lattice_cell                = self._structure.cell.array

        self.species                     = species

        self.num_atomic_layers           = num_atomic_layer_along_Z
        self.dist_from_surf              = dist_from_surf

        # self.adsorbate_shift = {'bridge': 0.0,
        #                         'ontop': 0.35,
        #                         'hollow': -0.2} # shift along z. Unit: angstrom

        self._Frac_coords                = self._structure.get_scaled_positions()
        self._Cart_coords                = self._structure.get_positions()
        max_z                            = np.max(self._Cart_coords, axis=0)[2]
        min_z                            = np.min(self._Cart_coords, axis=0)[2]
        layer_spacing                    = (max_z - min_z) / (num_atomic_layer_along_Z - 1)
        z_bottom                         = max_z - layer_spacing / 2
        self._top_layer_index            = np.where(self._Cart_coords[:,2] > z_bottom)

        self._top_layer_frac_coords      = self._Frac_coords[self._top_layer_index]
        self._top_layer_cart_coords      = self.fractional2cartesian(self.lattice_cell,
                                                                     self._Frac_coords[self._top_layer_index])

        self.crystal_type                = crystal_type
        crystal_density                  = {'fcc'  : 0.7404804896930611,
                                            'bcc'  : 0.6801747615878315,
                                            'hcp'  : 0.7404804896930611,
                                            'cubic': 0.5235987755982988}
        self.atomic_density              = crystal_density[self.crystal_type]
        self.atomic_diameter             = self.get_atomic_diameter_for_slab(layer_spacing,
                                                                             self.num_atomic_layers,
                                                                             self._structure)

        self.one_NN_top                  = self.get_1NN_of_top_layer(self._structure)

        # self.recommend_dist_from_surface = {
        #                                      'NiCoFeAlTi': { 'O': 1.4, 'OOH': 0.0}
        #                                    }

    def shift_fcoords(self, fcoord1, fcoord2):
        diff        = fcoord1 - fcoord2
        transition  = np.where(diff >= 0.5, 1.0, 0.0)
        fcoord2_new = fcoord2 + transition
        transition  = np.where(diff < -0.5, 1.0, 0.0)
        fcoord1_new = fcoord1 + transition
        return fcoord1_new, fcoord2_new

    def get_middle_fcoord(self, fcoords):
        fcoords_tmp = fcoords.copy()
        num_coord = np.shape(fcoords_tmp)[0]
        for i in range(num_coord):
            for j in range(i + 1, num_coord):
                fcoords_tmp[i], fcoords_tmp[j] =\
                self.shift_fcoords(fcoords_tmp[i], fcoords_tmp[j])
        return np.sum(fcoords_tmp, axis=0) / num_coord

    def get_1NN_cutoff(self):
        node_dict = {'fcc'  : (1.0 + 1.4148706415278178) / 2,
                     'bcc'  : (1.0 + 1.1547005383792515) / 2,
                     'hcp'  : (1.0 + 1.4148706415278178) / 2,
                     'cubic': (1.0 + 1.4142135623730951) / 2}
        return node_dict[self.crystal_type] * self.atomic_diameter

    def get_atomic_diameter_for_slab(self, layer_sapcing, atomic_layers, structure):
        cell_volume   = structure.get_volume()
        num_sites     = len(structure)
        effect_volume = layer_sapcing / self.length_z * atomic_layers * cell_volume
        return (effect_volume * self.atomic_density / num_sites * 3 / 4 / np.pi) ** (1 / 3) * 2

    def get_1NN_of_top_layer(self, structure=None):
        atoms_new = self._structure.copy()
        del atoms_new[:]
        for ccoord in self._top_layer_cart_coords:
            atoms_new.append(Atom('H', ccoord))
        self.dist_mat = atoms_new.get_all_distances(mic=True)
        OneNN_all = np.array(np.where(self.dist_mat < self.get_1NN_cutoff())).T
        return OneNN_all[np.where(OneNN_all[:,0] < OneNN_all[:,1])]

    def get_ontop_sites(self, x_shift=0.0, y_shift=0.0):
        ontop_coords_z = self._top_layer_cart_coords[:,2]
        ccoords = np.hstack((self._top_layer_cart_coords[:,[0,1]], np.reshape(ontop_coords_z, (len(ontop_coords_z), -1))))
        ccoords[:,0] += x_shift
        ccoords[:,1] += y_shift
        return ccoords, np.array([None for x in ccoords])

    def get_bridge_sites(self, x_shift=0.0, y_shift=0.0):
        fcoords  = []
        for i in self.one_NN_top:
            mid     = self.get_middle_fcoord(self._top_layer_frac_coords[[i[0], i[1]],:])
            fcoords.append(mid)
        fcoords = np.array(fcoords)
        ccoords = self.fractional2cartesian(self.lattice_cell, fcoords)
        ccoords[:,0] += x_shift
        ccoords[:,1] += y_shift
        return ccoords, np.array([None for x in ccoords])

    def get_disigma_sites(self, x_shift=0.0, y_shift=0.0):
        src, dst = [], [] # src: source; dst: destination
        for i in self.one_NN_top:
            fcoord1, fcoord2 = self._top_layer_frac_coords[[i[0], i[1]],:]
            src_tmp, dst_tmp = self.shift_fcoords(fcoord1, fcoord2)
            src.append(src_tmp)
            dst.append(dst_tmp)
        src, dst = np.array(src), np.array(dst)
        src = self.fractional2cartesian(self.lattice_cell, src)
        dst = self.fractional2cartesian(self.lattice_cell, dst)
        vectors = dst - src
        src[:,0] += x_shift
        src[:,1] += y_shift
        return src, vectors

    def get_hollow_sites(self, x_shift=0.0, y_shift=0.0):
        one_NN_top = self.one_NN_top.tolist()
        num_1nn    = len(one_NN_top)
        fcoord     = []
        for i in range(num_1nn):
            for j in range(i + 1, num_1nn):
                for k in range(j + 1, num_1nn):
                    nn_set = list(set(one_NN_top[i] + one_NN_top[j] + one_NN_top[k]))
                    if len(nn_set) == 3:
                        mid = self.get_middle_fcoord(self._top_layer_frac_coords[[nn_set[0], nn_set[1], nn_set[2]],:])
                        fcoord.append(mid)
        fcoords = np.array(fcoord)
        ccoords = self.fractional2cartesian(self.lattice_cell, fcoords)
        ccoords[:,0] += x_shift
        ccoords[:,1] += y_shift
        return ccoords, np.array([None for x in ccoords])

    def fractional2cartesian(self, vector_tmp, D_coord_tmp):
        C_coord_tmp = np.dot(D_coord_tmp, vector_tmp)
        return C_coord_tmp

    def cartesian2fractional(self, vector_tmp, C_coord_tmp):
        vector_tmp = np.mat(vector_tmp)
        D_coord_tmp = np.dot(C_coord_tmp, vector_tmp.I)
        D_coord_tmp = np.array(D_coord_tmp, dtype=float)
        return D_coord_tmp

    # def get_angle(self, vector1, vector2):
    #     v1_u = vector1 / np.linalg.norm(vector1)
    #     v2_u = vector2 / np.linalg.norm(vector2)
    #     return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

    def write_file_with_adsorption_sites(self, adsorbate_poscar, calculation_index=0):
        site_func = dict(zip(self._support_sites, [self.get_ontop_sites, self.get_bridge_sites, self.get_hollow_sites, self.get_disigma_sites]))

        if 'C3H6_pi' == self.species:
            x_shift=0.72
            y_shift=0.0
        else:
            x_shift=0.0
            y_shift=0.0

        cart_coords, site_str, out_array=[], [], []
        for site in self.sites:
            coord_tmp, out = site_func[site](x_shift=x_shift, y_shift=y_shift)
            cart_coords.append(coord_tmp)
            site_str += [site for x in coord_tmp]
            out_array += out.tolist()
        cart_coords = np.vstack(tuple(cart_coords))

        num_sites = np.shape(cart_coords)[0]
        for site_i in range(num_sites):
            f         = StringIO(adsorbate_poscar[self.species])
            adsorbate = read(f, format='vasp')
            surface   = self._structure.copy()
            position  = (cart_coords[site_i][0], cart_coords[site_i][1])

            if site_str[site_i] == 'disigma':
                src_v = adsorbate.arrays['positions'][0] - adsorbate.arrays['positions'][1]
                src_v[2] = 0.0
                dst_v = out_array[site_i]
                dst_v[2] = 0.0
                adsorbate.rotate(src_v, dst_v, center ='COP')
                displacement = np.array([[0.0, 0.0, 0.0] for x in surface])

            add_adsorbate(surface, adsorbate, self.dist_from_surf, position)
            if site_str[site_i] == 'disigma':
                displacement_tmp = dst_v / np.linalg.norm(dst_v) * -0.72
                displacement = np.vstack([displacement, [displacement_tmp for x in range(len(adsorbate))]])
                surface.translate(displacement)
            write(f'POSCAR_{calculation_index}_{site_i}', surface, format='vasp')
        return num_sites

if __name__ == '__main__':
    adder = AddAtoms('POSCAR_surf_opt_1.gat',   # input file name.
                     species='C3H6_di',               # name of adsorbates. O, H, OH, OOH, C3H6_di, C3H6_pi, C3H7
                     sites='disigma',            # name of sites. ontop, bridge, hollow, disigma, or all
                     dist_from_surf=1.7,        # distance between adsorbate and the surface. Decrease this variable if the adsorbate is too high.
                     num_atomic_layer_along_Z=6) # number of atomic layers along z direction of the slab model.
    from agat.lib.adsorbate_poscar import adsorbate_poscar
    num_sites = adder.write_file_with_adsorption_sites(adsorbate_poscar)



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 24 22:01:12 2023

@author: ZHANG Jun
"""

import os
import json
import copy

import torch

def save_model(model, model_save_dir='agat_model'):
    """Saving PyTorch model to the disk. Save PyTorch model, including parameters and structure. See: https://pytorch.org/tutorials/beginner/basics/saveloadrun_tutorial.html

    :param model: A PyTorch-based model.
    :type model: PyTorch-based model.
    :param model_save_dir: A directory to store the model, defaults to 'agat_model'
    :type model_save_dir: str, optional
    :output: A file saved to the disk under ``model_save_dir``.
    :outputtype: A file.

    """
    if not os.path.exists(model_save_dir):
        os.makedirs(model_save_dir)
    torch.save(model, os.path.join(model_save_dir, 'agat.pth'))

def load_model(model_save_dir='agat_model', device='cuda'):
    """Loading PyTorch model from the disk.

    :param model_save_dir: A directory to store the model, defaults to 'agat_model'
    :type model_save_dir: str, optional
    :param device: Device for the loaded model, defaults to 'cuda'
    :type device: str, optional
    :return: A PyTorch-based model.
    :rtype: PyTorch-based model.

    """
    device = torch.device(device)
    if device.type == 'cuda':
        new_model = torch.load(os.path.join(model_save_dir, 'agat.pth'))
    elif device.type == 'cpu':
        new_model = torch.load(os.path.join(model_save_dir, 'agat.pth'),
                               map_location=torch.device(device))
    new_model.eval()
    new_model = new_model.to(device)
    new_model.device = device
    return new_model

def save_state_dict(model, state_dict_save_dir='agat_model', **kwargs):
    """Saving state dict (model weigths and other input info) to the disk. See: https://pytorch.org/tutorials/beginner/basics/saveloadrun_tutorial.html

    :param model: A PyTorch-based model.
    :type model: PyTorch-based model.
    :param state_dict_save_dir: A directory to store the model state dict (model weigths and other input info), defaults to 'agat_model'
    :type state_dict_save_dir: str, optional
    :param **kwargs: More information you want to save.
    :type **kwargs: kwargs
    :output: A file saved to the disk under ``model_save_dir``.
    :outputtype: A file

    """
    if not os.path.exists(state_dict_save_dir):
        os.makedirs(state_dict_save_dir)
    checkpoint_dict = {**{'model_state_dict': model.state_dict()}, **kwargs}
    torch.save(checkpoint_dict, os.path.join(state_dict_save_dir, 'agat_state_dict.pth'))

def load_state_dict(state_dict_save_dir='agat_model'):
    """Loading state dict (model weigths and other info) from the disk. See: https://pytorch.org/tutorials/beginner/basics/saveloadrun_tutorial.html

    :param state_dict_save_dir:  A directory to store the model state dict (model weigths and other info), defaults to 'agat_model'
    :type state_dict_save_dir: str, optional
    :return: State dict.
    :rtype: TYPE

    .. note::
        Reconstruct a model/optimizer before using the loaded state dict.

        Example::

            model = PotentialModel(...)
            model.load_state_dict(checkpoint['model_state_dict'])
            new_model.eval()
            model = model.to(device)
            model.device = device
            optimizer = ...
            optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    """

    checkpoint_dict = torch.load(os.path.join(state_dict_save_dir, 'agat_state_dict.pth'))
    return checkpoint_dict

def config_parser(config):
    """Parse the input configurations/settings.

    :param config: configurations
    :type config: str/dict. if str, load from the json file.
    :raises TypeError: DESCRIPTION
    :return: TypeError('Wrong configuration type.')
    :rtype: TypeError

    """

    if isinstance(config, dict):
        return config
    elif isinstance(config, str):
        with open(config, 'r') as config_f:
            return json.load(config_f)
    elif isinstance(config, type(None)):
        return {}
    else:
        raise TypeError('Wrong configuration type.')

class EarlyStopping:
    def __init__(self, model, logger, patience=10, model_save_dir='model_save_dir'):
        """Stop training when model performance stop improving after some steps.

        :param model: AGAT model
        :type model: torch.nn
        :param logger: I/O file
        :type logger: _io.TextIOWrapper
        :param patience: Stop patience, defaults to 10
        :type patience: int, optional
        :param model_save_dir: A directory to save the model, defaults to 'model_save_dir'
        :type model_save_dir: str, optional


        """

        self.model      = model
        self.patience   = patience
        self.counter    = 0
        self.best_score = None
        self.update     = None
        self.early_stop = False
        self.logger     = logger
        self.model_save_dir = model_save_dir

        if not os.path.exists(self.model_save_dir):
            os.mkdir(model_save_dir)
        self.save_model_info()

    def step(self, score, epoch, model, optimizer):
        if self.best_score is None:
            self.best_score = score
            self.update = True
            # self.save_model(model, model_save_dir=self.model_save_dir)
            # self.save_checkpoint(model, model_save_dir=self.model_save_dir)
        elif score > self.best_score:
            self.update = False
            self.counter += 1
            print(f'User log: EarlyStopping counter: {self.counter} out of {self.patience}',
                  file=self.logger)
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.update     = True
            self.save_model(model)
            save_state_dict(model, state_dict_save_dir=self.model_save_dir,
                            optimizer_state_dict=optimizer.state_dict(),
                            epoch=epoch, total_loss=score)
            self.counter = 0
        return self.early_stop

    def save_model(self, model):
        '''Saves model when validation loss decrease.'''
        torch.save(model, os.path.join(self.model_save_dir, 'agat.pth'))
        print(f'User info: Save model with the best score: {self.best_score}',
              file=self.logger)

    # def save_checkpoint(self, model):
    #     '''Saves model when validation loss decrease.'''
    #     torch.save({model.state_dict()}, os.path.join(self.model_save_dir, 'agat_state_dict.pth'))

    def save_model_info(self):
        info = copy.deepcopy(self.model.__dict__)
        info = {k:v for k,v in info.items() if isinstance(v, (str, list, int, float))}
        with open(os.path.join(self.model_save_dir, 'agat_model.json'), 'w') as f:
            json.dump(info, f, indent=4)

def load_graph_build_method(path):
    """ Load graph building scheme. This file is normally saved when you build your dataset.

    :param path: Directory for storing ``graph_build_scheme.json`` file.
    :type path: str
    :return: A dict denotes how to build the graph.
    :rtype: dict

    """

    json_file  = path

    assert os.path.exists(json_file), f"{json_file} file dose not exist."
    with open(json_file, 'r') as jsonf:
        graph_build_scheme = json.load(jsonf)
    return graph_build_scheme

def PearsonR(y_true, y_pred):
    """Calculating the Pearson coefficient.

    :param y_true: The first torch.tensor.
    :type y_true: torch.Tensor
    :param y_pred: The second torch.tensor.
    :type y_pred: torch.Tensor
    :return: Pearson coefficient
    :rtype: torch.Tensor

    .. Note::

        It looks like the `torch.jit.script` decorator is not helping in comuputing large `torch.tensor`, see `agat/test/tesor_computation_test.py` for more details.

    """
    ave_y_true = torch.mean(y_true)
    ave_y_pred = torch.mean(y_pred)

    y_true_diff = y_true - ave_y_true
    y_pred_diff = y_pred - ave_y_pred

    above = torch.sum(torch.mul(y_true_diff, y_pred_diff))
    below = torch.mul(torch.sqrt(torch.sum(torch.square(y_true_diff))),
                             torch.sqrt(torch.sum(torch.square(y_pred_diff))))
    return torch.divide(above, below)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 11 13:22:58 2023

@author: ZHANG Jun
"""

from distutils.core import setup
from setuptools import find_packages

with open("README.md", "r") as f:
  long_description = f.read()

# print(long_description)
# with open('requirements.txt', "r") as f:
#     requirements = f.readlines()
#     requirements = [x.strip() for x in requirements]

# long_description = (this_directory / "README.md").read_text()

setup(name='agat',
      version='8.0.5',
      python_requires='>=3.10',
      description='Atomic Graph ATtention networks for predicting atomic energies and forces.',
      long_description=long_description,
      include_package_data=True,
      long_description_content_type='text/markdown',
      author='ZHANG Jun; ZHAO Shijun',
      author_email='j.zhang@my.cityu.edu.hk',
      url='https://github.com/jzhang-github/AGAT',
      install_requires=['numpy',
                        'ase',
                        'tqdm'],
      license='GPL',
      packages=find_packages(exclude=['cata_old', 'tools']),
      platforms=["all"],
      classifiers=[
                # How mature is this project? Common values are
                #   3 - Alpha
                #   4 - Beta
                #   5 - Production/Stable
                'Development Status :: 3 - Alpha',
                'License :: OSI Approved :: GNU General Public License (GPL)',
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.10',
          'Topic :: Software Development :: Libraries'
      ],
      )



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 11:04:06 2023

@author: ZHANG Jun
"""

import torch
import torch.nn as nn
from dgl import function as fn
from dgl.ops import edge_softmax

class Layer(nn.Module):
    """
    Description:
    ----------
        Single graph attention network for crystal.
    Parameters
    ----------
    in_dim: int
        Depth of node representation in the input of each head of this `GAT` layer.
    out_dim: int
        Depth of node representation in the output of each head of this `GAT` layer.
    bias: Boolean
        Whether the dense layer uses a bias vector.
    negative_slope: float
        Negative slope coefficient


    Meaning of abbreviations:
    ----------
    dist:
        distance matrix
    feat:
        features
    ft:
        features
    src:
        source node
    dst:
        destination node
    e:
        raw e_i_j: refer to: https://arxiv.org/abs/1710.10903
    a:
        alpha_i_j: refer to: https://arxiv.org/abs/1710.10903
    att:
        attention mechanism
	act:
        activation function
    """

    def __init__(self,
                 in_dim,
                 out_dim,
                 num_heads,
                 device='cuda',
                 bias=True,
                 negative_slope=0.2):
        super(Layer, self).__init__()
        # input args
        self.in_dim         = in_dim
        self.out_dim        = out_dim
        self.num_heads      = num_heads
        self.device         = device
        self.bias           = bias
        self.negative_slope = negative_slope

        # initialize trainable parameters
        self.w_att_src = nn.Parameter(torch.randn(1, self.num_heads, self.out_dim, device=self.device))
        self.w_att_dst = nn.Parameter(torch.randn(1, self.num_heads, self.out_dim, device=self.device))

        # dense layers
        self.layer = nn.Linear(self.in_dim, self.out_dim*self.num_heads, bias=self.bias, device=self.device)
        self.leaky_relu1 = nn.LeakyReLU(negative_slope=self.negative_slope)

        # leaky relu function
        self.leaky_relu2 = nn.LeakyReLU(negative_slope=self.negative_slope)

    def forward(self, feat, dist, graph): # feat with shape of (number of nodes, number of features of each node). The graph can have no values of nodes.
        """
        Description:
        ----------
            Forward this `GAT` layer.
        Parameters
        ----------
        feat: `torch.Tensor`
            Input features of all nodes (atoms).
        graph: `DGLGraph`
            A graph built with DGL.
        """
        feat_src = feat_dst = self.leaky_relu1(self.layer(feat))

        # attention
        feat_src = torch.reshape(feat_src, (-1, self.num_heads, self.out_dim)) # shape: (number of nodes, number of heads, num out)
        feat_dst = torch.reshape(feat_dst, (-1, self.num_heads, self.out_dim)) # shape: (number of nodes, number of heads, num out)

        # shape of e_src and e_dst after `torch.sum`: (number of nodes, number of heads, 1)
        e_src = torch.sum(feat_src * self.w_att_src, axis=-1, keepdim=True)
        e_dst = torch.sum(feat_dst * self.w_att_dst, axis=-1, keepdim=True)

        # save on nodes
        graph.srcdata.update({'ft': feat_src, 'e_src': e_src})
        graph.dstdata.update({'e_dst': e_dst})

        # save on edges
        graph.apply_edges(fn.u_add_v('e_src', 'e_dst', 'e'))   # shape of e: (number of edges, number of heads, 1) # similar to the original paper, but feed to dense first, summation is the second.
        e = self.leaky_relu2(graph.edata.pop('e'))  # shape of e: (number of edges, number of heads, 1)

        dist = torch.reshape(dist, (-1, self.num_heads, 1))
        graph.edata['a']  = edge_softmax(graph, e) * dist # shape of a: (number of edges, number of heads, 1) # shape of dist: (number of edges, number of heads, 1)

        graph.update_all(fn.u_mul_e('ft', 'a', 'm'), fn.sum('m', 'ft')) # shape of ft: (number of nodes, number of heads, number of out)
        dst = torch.reshape(graph.ndata['ft'], (-1, self.num_heads * self.out_dim)) # shape of `en`: (number of nodes, number of heads * number of out)
        return dst # node_energy, node_force

if __name__ == '__main__':
    import dgl
    g_list, l_list = dgl.load_graphs('all_graphs.bin')
    graph = g_list[1].to('cuda')
    feat = graph.ndata['h']
    dist = graph.edata['dist']

    GL = Layer(6, 10, 2, device='cuda')
    out = GL.forward(feat, dist, graph)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep  3 17:16:24 2021

@author: ZHANG Jun
"""

import json
import os
import sys

import torch
from ase.calculators.calculator import Calculator

from ..data.build_dataset import CrystalGraph
from ..lib.model_lib import load_model

class AgatCalculator(Calculator):
    """Calculator with ASE module.

    :param model_save_dir: Directory storing the well-trained model.
    :type model_save_dir: str
    :param graph_build_scheme_dir: Direcotry storing the ``graph_build_scheme.json`` file.
    :type graph_build_scheme_dir: str
    :param device: model device, defaults to 'cuda'
    :type device: str, optional
    :param **kwargs: other input arguments
    :type **kwargs: dict
    :return: Calculated properties.
    :rtype: dict

    Example::

        model_save_dir = 'agat_model'
        graph_build_scheme_dir = 'dataset'
        atoms = read('CONTCAR')
        calculator=AgatCalculator(model_save_dir,
                                  graph_build_scheme_dir)
        atoms = Atoms(atoms, calculator=calculator)
        dyn = BFGS(atoms, trajectory='test.traj')
        dyn.run(fmax=0.005)

        traj = read('test.traj', index=':')
        write("XDATCAR.gat", traj)

    """
    implemented_properties = ['energy', 'forces']
    default_parameters  = { }
    ignored_changes = set()
    def __init__(self, model_save_dir, graph_build_scheme_dir, device = 'cuda',
                 **kwargs):
        Calculator.__init__(self, **kwargs)
        # self.atoms = None  # copy of atoms object from last calculation
        # self.results = {}  # calculated properties (energy, forces, ...)
        # self.parameters = None  # calculational parameters
        # self._directory = None  # Initialize

        self.model_save_dir = model_save_dir
        self.graph_build_scheme_dir = graph_build_scheme_dir
        self.device = device

        self.model = load_model(self.model_save_dir, self.device)
        self.graph_build_scheme = self.load_graph_build_scheme(self.graph_build_scheme_dir)

        build_properties = {'energy': False, 'forces': False, 'cell': False,
                            'cart_coords': False, 'frac_coords': False, 'path': False,
                            'stress': False} # We only need the topology connections.
        self.graph_build_scheme['build_properties'] = {**self.graph_build_scheme['build_properties'],
                                                       **build_properties}
        self.cg = CrystalGraph(**self.graph_build_scheme)

    # def set(self):
    #     pass

    def load_graph_build_scheme(self, path):
        """ Load graph building scheme. This file is normally saved when you build your dataset.

        :param path: Directory for storing ``graph_build_scheme.json`` file.
        :type path: str
        :return: A dict denotes how to build the graph.
        :rtype: dict

        """
        json_file  = os.path.join(path, 'graph_build_scheme.json')
        assert os.path.exists(json_file), f"{json_file} file dose not exist."
        with open(json_file, 'r') as jsonf:
            graph_build_scheme = json.load(jsonf)
        return graph_build_scheme

    def calculate(self, atoms=None, properties=None, system_changes=['positions', 'numbers', 'cell', 'pbc']):
        """

        :param atoms: ase.atoms object, defaults to None
        :type atoms: ase.atoms, optional
        :param properties: calculated properties, defaults to None
        :type properties: none, optional
        :param system_changes: DESCRIPTION, defaults to ['positions', 'numbers', 'cell', 'pbc']
        :type system_changes: TYPE, optional
        :return: calculated results
        :rtype: dict

        """

        if properties is None:
            properties = self.implemented_properties

        if atoms is not None:
            self.atoms = atoms.copy()

        # read graph
        graph, info = self.cg.get_graph(atoms)
        graph = graph.to(self.device)

        with torch.no_grad():
            energy_pred, force_pred, stress_pred = self.model.forward(graph)

        self.results = {'energy': energy_pred[0].item() * len(atoms),
                        'forces': force_pred.cpu().numpy()}

# class GatCalculator(Calculator):
#     """ Calculator with ASE module. Pymatgen is also needed.

#     :param energy_model_save_path: Directory for the saved energy model.
#     :type energy_model_save_path: str
#     :param force_model_save_path: Directory for the saved force model.
#     :type force_model_save_path: str
#     :param graph_build_scheme_path: Directory for the json file of how to build graphs. This file is normally saved when you build your dataset.
#     :type graph_build_scheme_path: str
#     :param load_model: Load well-trained models or not, defaults to True
#     :type load_model: bool, optional
#     :param **kwargs: kwargs
#     :type **kwargs: kwargs

#     Example::

#         energy_model_save_dir = os.path.join('out_file', 'energy_ckpt')
#         force_model_save_dir = os.path.join('out_file', 'force_ckpt')
#         graph_build_scheme_dir = 'dataset'
#         poscar = read('POSCAR')
#         calculator=GatCalculator(energy_model_save_dir,
#                                     force_model_save_dir,
#                                     graph_build_scheme_dir)
#         poscar = Atoms(poscar, calculator=calculator)
#         dyn = BFGS(poscar, trajectory='test.traj')
#         dyn.run(fmax=0.005)

#         traj = read('test.traj', index=':')
#         write("XDATCAR.gat", traj)

#     """
#     implemented_properties = ['energy', 'energies', 'free_energy', 'forces', 'stress', 'stresses']
#     default_parameters     = { }
#     def __init__(self,
#                  energy_model_save_path,
#                  force_model_save_path,
#                  graph_build_scheme_path,
#                  load_model=True, **kwargs):


#         Calculator.__init__(self, **kwargs)
#         self.app = GatApp(energy_model_save_path,
#                           force_model_save_path,
#                           graph_build_scheme_path,
#                           load_model=True)


#     def calculate(self, atoms=None, properties=None, system_changes=['positions', 'numbers', 'cell', 'pbc']):
#         if properties is None:
#             properties = self.implemented_properties

#         # save `atom` object to a file or to the memory
#         write('CONTCAR.gat', atoms)

#         # read file with MP
#         graph, info = self.app.get_graph('CONTCAR.gat')
#         # graph = graph.to(self.app.device)
#         # print(graph.device)

#         # get results
#         energy = self.app.get_energy(graph)
#         forces = self.app.get_forces(graph)

#         self.results = {'energy': energy,
#                         'forces': forces} # ,                        'energies': energies

# debug
if __name__ == '__main__':
    pass

